"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numbers
import graphviz
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import semopy
import subprocess

from matplotlib import colors as mcolors
from matplotlib.colors import is_color_like
from sklearn import linear_model
from sklearn.linear_model import LassoLarsIC, LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array, check_scalar

from ._rcd import extract_ancestors
from ._f_correlation import f_correlation
from ._visualize_nonlinear_causal_effect import visualize_nonlinear_causal_effect
from ._common_edge import get_common_edge_probabilities, print_common_edge_directions, make_dot_for_nan_probability_matrix

__all__ = [
    "print_causal_directions",
    "print_dagc",
    "make_prior_knowledge",
    "remove_effect",
    "make_dot",
    "make_dot_highlight",
    "predict_adaptive_lasso",
    "get_sink_variables",
    "get_exo_variables",
    "find_all_paths",
    "likelihood_i",
    "log_p_super_gaussian",
    "variance_i",
    "extract_ancestors",
    "f_correlation",
    "visualize_nonlinear_causal_effect",
    "evaluate_model_fit",
    "calculate_distance_from_root_nodes",
    "calculate_total_effect",
    "get_common_edge_probabilities",
    "print_common_edge_directions",
    "make_dot_for_nan_probability_matrix",
]


def print_causal_directions(cdc, n_sampling, labels=None):
    """Print causal directions of bootstrap result to stdout.

    Parameters
    ----------
    cdc : dict
        List of causal directions sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_causal_direction_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (fr, to, co) in enumerate(zip(cdc["from"], cdc["to"], cdc["count"])):
        sign = "" if "sign" not in cdc else "(b>0)" if cdc["sign"][i] > 0 else "(b<0)"
        if labels:
            print(f"{labels[to]} <--- {labels[fr]} {sign} ({100*co/n_sampling:.1f}%)")
        else:
            print(f"x{to} <--- x{fr} {sign} ({100*co/n_sampling:.1f}%)")


def print_dagc(dagc, n_sampling, labels=None):
    """Print DAGs of bootstrap result to stdout.

    Parameters
    ----------
    dagc : dict
        List of directed acyclic graphs sorted by count in descending order.
        This can be set the value returned by ``BootstrapResult.get_directed_acyclic_graph_counts()`` method.
    n_sampling : int
        Number of bootstrapping samples.
    labels : array-like, optional (default=None)
        List of feature lables.
        If set labels, the output feature name will be the specified label.
    """
    for i, (dag, co) in enumerate(zip(dagc["dag"], dagc["count"])):
        print(f"DAG[{i}]: {100*co/n_sampling:.1f}%")
        for j, (fr, to) in enumerate(zip(dag["from"], dag["to"])):
            sign = (
                "" if "sign" not in dag else "(b>0)" if dag["sign"][j] > 0 else "(b<0)"
            )
            if labels:
                print("\t" + f"{labels[to]} <--- {labels[fr]} {sign}")
            else:
                print("\t" + f"x{to} <--- x{fr} {sign}")


def make_prior_knowledge(
    n_variables,
    exogenous_variables=None,
    sink_variables=None,
    paths=None,
    no_paths=None,
):
    """Make matrix of prior knowledge.

    Parameters
    ----------
    n_variables : int
        Number of variables.
    exogenous_variables : array-like, shape (index, ...), optional (default=None)
        List of exogenous variables(index).
        Prior knowledge is created with the specified variables as exogenous variables.
    sink_variables : array-like, shape (index, ...), optional (default=None)
        List of sink variables(index).
        Prior knowledge is created with the specified variables as sink variables.
    paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs with directed path.
        If ``(i, j)``, prior knowledge is created that xi has a directed path to xj.
    no_paths : array-like, shape ((index, index), ...), optional (default=None)
        List of variables(index) pairs without directed path.
        If ``(i, j)``, prior knowledge is created that xi does not have a directed path to xj.

    Returns
    -------
    prior_knowledge : array-like, shape (n_variables, n_variables)
        Return matrix of prior knowledge used for causal discovery.
    """
    prior_knowledge = np.full((n_variables, n_variables), -1)
    if no_paths:
        for no_path in no_paths:
            prior_knowledge[no_path[1], no_path[0]] = 0
    if paths:
        for path in paths:
            prior_knowledge[path[1], path[0]] = 1
    if sink_variables:
        for var in sink_variables:
            prior_knowledge[:, var] = 0
    if exogenous_variables:
        for var in exogenous_variables:
            prior_knowledge[var, :] = 0
    np.fill_diagonal(prior_knowledge, -1)
    return prior_knowledge


def get_sink_variables(adjacency_matrix):
    """The sink variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    sink_variables : array-like
        List of sink variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    sink_vars = [i for i in range(am.shape[1]) if am[:, i].sum() == 0]
    return sink_vars


def get_exo_variables(adjacency_matrix):
    """The exogenous variables(index) in the adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_variables, n_variables)
        Adjacency matrix, where n_variables is the number of variables.

    Returns
    -------
    exogenous_variables : array-like
        List of exogenous variables(index).
    """
    am = adjacency_matrix.copy()
    am = np.abs(am)
    np.fill_diagonal(am, 0)
    exo_vars = [i for i in range(am.shape[1]) if am[i, :].sum() == 0]
    return exo_vars


def remove_effect(X, remove_features, return_coefs=False):
    """Create a dataset that removes the effects of features by linear regression.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    remove_features : array-like, shape (n_removes,)
        List of features(index) to remove effects.
    return_coefs : bool, optional (default=False)
        Return regression coefficients or not.

    Returns
    -------
    X : array-like, shape (n_samples, n_features)
        Data after removing effects of ``remove_features``.
    coefs : dict, optional
        Coefficients estimated by linear regression.
        The keys are indices of remaining features and
        the values are lists of coefficients of removed features.
        The order of the coefficients in the list is the same
        as in remove_features. Only provided if return_coefs is True.
    """
    X = np.copy(check_array(X))
    coefs = {}
    features_ = [i for i in np.arange(X.shape[1]) if i not in remove_features]
    for feature in features_:
        reg = linear_model.LinearRegression()
        reg.fit(X[:, remove_features], X[:, feature])
        X[:, feature] = X[:, feature] - reg.predict(X[:, remove_features])
        coefs[feature] = reg.coef_
    if return_coefs:
        return X, coefs
    return X


def make_dot(
    adjacency_matrix,
    labels=None,
    lower_limit=0.01,
    prediction_feature_indices=None,
    prediction_target_label="Y(pred)",
    prediction_line_color="red",
    prediction_coefs=None,
    prediction_feature_importance=None,
    path=None,
    path_color=None,
    detect_cycle=False,
    ignore_shape=False,
):
    """Directed graph source code in the DOT language with specified adjacency matrix.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Adjacency matrix to make graph, where ``n_features`` is the number of features.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    prediction_feature_indices : array-like, optional (default=None)
        Indices to use as prediction features.
    prediction_target_label : string, optional (default='Y(pred)'))
        Label to use for target variable of prediction.
    prediction_line_color : string, optional (default='red')
        Line color to use for prediction's graph.
    prediction_coefs : array-like, optional (default=None)
        Coefficients to use for prediction's graph.
    prediction_feature_importance : array-like, optional (default=None)
        Feature importance to use for prediction's graph.
    path : tuple, optional (default=None)
        Path to highlight. Tuple of start index and end index.
    path_color : string, optional (default=None)
        Colors to highlight a path.
    detect_cycle : boolean, optional (default=False)
        Highlight simple cycles.
    ignore_shape : boolean, optional (default=False)
        Ignore checking the shape of adjaceny_matrix or not.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """
    # Check parameters
    B = check_array(np.nan_to_num(adjacency_matrix))
    if not ignore_shape and B.shape[0] != B.shape[1]:
        raise ValueError("'adjacency_matrix' is not square matrix.")
    if labels is not None:
        if B.shape[1] != len(labels):
            raise ValueError(
                "Length of 'labels' does not match length of 'adjacency_matrix'"
            )
    if prediction_feature_indices is not None:
        if prediction_coefs is not None and (
            len(prediction_feature_indices) != len(prediction_coefs)
        ):
            raise ValueError(
                "Length of 'prediction_coefs' does not match length of 'prediction_feature_indices'"
            )
        if prediction_feature_importance is not None and (
            len(prediction_feature_indices) != len(prediction_feature_importance)
        ):
            raise ValueError(
                "Length of 'prediction_feature_importance' does not match length of 'prediction_feature_indices'"
            )

    if path is not None:
        if not isinstance(path, tuple) or len(path) < 2:
            raise TypeError("'path' should be a tuple of node indices.")
        if path_color is not None and not is_color_like(path_color):
            raise ValueError("'path_color' should be an color name.")

        if path_color is None:
            path_color = "black"
        path_start = check_scalar(
            path[0], "path_start", int, min_val=0, max_val=B.shape[0]
        )
        path_end = check_scalar(path[1], "path_end", int, min_val=0, max_val=B.shape[0])

    if detect_cycle is not None:
        if not isinstance(detect_cycle, bool):
            raise TypeError("'detect_cycle' should be boolean.")

    idx = np.abs(B) > lower_limit

    nx_graph = None
    if path is not None or detect_cycle is True:
        nx_graph = nx.from_numpy_array(idx.astype(int).T, create_using=nx.DiGraph)

    # search path
    path_edges = None
    path_nodes = None
    if path is not None:
        path_edges = set()
        for path_ in nx.all_simple_paths(nx_graph, path[0], path[1]):
            path_edges |= set(zip(path_, path_[1:]))

        path_nodes = np.unique([list(edge) for edge in path_edges])
        if labels is None:
            path_nodes = [f"x{node}" for node in path_nodes]
        else:
            path_nodes = [labels[node] for node in path_nodes]

    # simple cycles
    cycle_edges = []
    if detect_cycle is True:
        for cycle in nx.simple_cycles(nx_graph):
            es = list(zip(cycle, np.roll(cycle, -1)))
            cycle_edges.append(es)
        cycle_edges = sum(cycle_edges, start=[])

    d = graphviz.Digraph(engine="dot")

    # nodes
    names = labels if labels else [f"x{i}" for i in range(len(B))]
    for name in names:
        kwargs = {}
        if path is not None:
            if name in path_nodes:
                kwargs["fontcolor"] = path_color
                kwargs["color"] = path_color
            if name == names[path_start] or name == names[path_end]:
                kwargs["peripheries"] = "2"
        d.node(name, **kwargs)

    # edges
    dirs = np.where(idx)
    for to, from_, coef in zip(dirs[0], dirs[1], B[idx]):
        kwargs = {}
        if path is not None and (from_, to) in path_edges:
            kwargs["penwidth"] = "2"
            kwargs["fontcolor"] = path_color
            kwargs["color"] = path_color
        if detect_cycle is True and (from_, to) in cycle_edges:
            kwargs["style"] = "dashed"
        d.edge(names[from_], names[to], label=f"{coef:.2f}", **kwargs)

    # integrate of prediction model
    if prediction_feature_indices is not None:
        d.node(
            prediction_target_label,
            color=prediction_line_color,
            fontcolor=prediction_line_color,
        )

        if prediction_coefs is not None:
            for from_, coef in zip(prediction_feature_indices, prediction_coefs):
                if np.abs(coef) > lower_limit:
                    d.edge(
                        names[from_],
                        prediction_target_label,
                        label=f"{coef:.2f}",
                        color=prediction_line_color,
                        fontcolor=prediction_line_color,
                        style="dashed",
                    )

        elif prediction_feature_importance is not None:
            for from_, imp in zip(
                prediction_feature_indices, prediction_feature_importance
            ):
                d.edge(
                    names[from_],
                    prediction_target_label,
                    label=f"({imp})",
                    color=prediction_line_color,
                    fontcolor=prediction_line_color,
                    style="dashed",
                )

        else:
            for from_ in prediction_feature_indices:
                d.edge(
                    names[from_],
                    prediction_target_label,
                    color=prediction_line_color,
                    style="dashed",
                )

    # If the value is nan, draw a double-headed arrow
    unk_order = np.where(np.isnan(np.tril(adjacency_matrix)))
    unk_order_set = set([val for item in unk_order for val in item])
    with d.subgraph() as s:
        s.attr(rank="same")
        for node in unk_order_set:
            s.node(names[node])
    for to, from_ in zip(unk_order[0], unk_order[1]):
        d.edge(names[from_], names[to], dir="both")

    return d


def make_dot_highlight(
    adjacency_matrix,
    node_index,
    labels=None,
    max_dsc=None,
    max_anc=None,
    lower_limit=None,
    cmap=None,
    draw_others=True,
    detect_cycle=False,
    path=None,
    path_color=None,
    vmargin=2,
    hmargin=2,
):
    """Hierachical make_dot.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Adjacency matrix to make graph, where ``n_features`` is the number of features.
    node_index : int
        Index of target node.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    max_dsc : int, optional (default=None)
        Number of hierarchies on the descendant side to display.
    max_asc : int, optional (default=None)
        Number of hierarchies on the ancestral side to display.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    cmap : colormap, optional (default=None)
        Gradient color to represent hierarchy.
    draw_others : boolean, optional (default=True)
        Whether to draw others cluster.
    detect_cycle : boolean, optional (default=False)
        Highlight simple cycles.
    path : tuple, optional (default=None)
        Path to highlight. Tuple of start index and end index.
    path_color : string, optional (default=None)
        Colors to highlight a path.
    vmargin : float, optional (default=2)
        Vertical margin between nodes.
    hmargin : float, optional (default=2)
        Horizontal margin between nodes.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """

    # margins for clusters
    cls_margin_l = 1.5
    cls_margin_r = 0.6
    cls_margin_t = 0.6
    cls_margin_b = 0.4
    cls_space_target_other = 0.2

    # check arguments
    adj = check_array(
        adjacency_matrix,
        ensure_2d=True,
        ensure_min_samples=2,
        ensure_min_features=2,
        copy=True,
    )
    if adj.shape[0] != adj.shape[1]:
        raise ValueError

    node_index = check_scalar(
        node_index,
        "node_index",
        int,
        min_val=0,
        max_val=adj.shape[0],
        include_boundaries="left",
    )

    if labels is None:
        node_names = [f"x{i}" for i in range(adj.shape[1])]
    else:
        node_names = check_array(labels, ensure_2d=False, dtype=str)
        if len(node_names) != adj.shape[1]:
            raise ValueError

    if max_dsc is not None:
        max_dsc = check_scalar(max_dsc, "max_dsc", int, min_val=0)

    if max_anc is not None:
        max_anc = check_scalar(max_anc, "max_anc", int, min_val=0)

    if lower_limit is not None:
        lower_limit = check_scalar(lower_limit, "lower_limit", (int, float))

    draw_others = check_scalar(draw_others, "draw_others", bool)
    detect_cycle = check_scalar(detect_cycle, "detect_cycle", bool)

    path_start = None
    path_end = None
    if path is not None:
        if not isinstance(path, tuple) or len(path) < 2:
            raise TypeError("'path' should be a tuple of node indices.")
        if path_color is not None and not is_color_like(path_color):
            raise ValueError("'path_color' should be an color name.")
        if path_color is None:
            path_color = "black"
        path_start = check_scalar(
            path[0], "start", int, min_val=0, max_val=adj.shape[0]
        )
        path_end = check_scalar(path[1], "end", int, min_val=0, max_val=adj.shape[0])

    vmargin = check_scalar(vmargin, "vmargin", (float, int), min_val=1)
    hmargin = check_scalar(hmargin, "hmargin", (float, int), min_val=1)

    # apply lower_limit
    if lower_limit is not None:
        adj[abs(adj) < lower_limit] = 0

    # analyze network
    G = nx.from_numpy_array(~np.isclose(adj.T, 0), create_using=nx.DiGraph)

    dsc_path_length = nx.single_source_shortest_path_length(
        G, node_index, cutoff=max_dsc
    )
    del dsc_path_length[node_index]

    anc_path_length = nx.single_source_shortest_path_length(
        G.reverse(), node_index, cutoff=max_anc
    )
    del anc_path_length[node_index]

    # node indices
    dsc_indices = set(dsc_path_length.keys())
    anc_indices = set(anc_path_length.keys())
    isolate_indices = set(nx.isolates(G))
    other_indices = (
        set(np.arange(adj.shape[0]))
        - dsc_indices
        - anc_indices
        - isolate_indices
        - set([node_index])
    )

    # clusters (distance -> list of nodes)
    clusters = {}

    clusters[0] = [node_index]

    clusters[None] = sorted(other_indices) + sorted(isolate_indices)

    for node in dsc_indices:
        d = dsc_path_length[node]
        if d not in clusters.keys():
            clusters[d] = []
        clusters[d].append(node)

    for node in anc_indices:
        d = -anc_path_length[node]
        if d not in clusters.keys():
            clusters[d] = []
        clusters[d].append(node)

    # search path
    path_edges = None
    path_nodes = None
    if path is not None:
        path_edges = set()
        for path_ in nx.all_simple_paths(G, path_start, path_end):
            path_edges |= set(zip(path_, path_[1:]))

        path_nodes = np.unique([list(edge) for edge in path_edges])
        path_nodes = [node_names[node] for node in path_nodes]

    # create graphviz graph
    graph_attr = {"rankdir": "TB", "splines": "true"}
    graph = graphviz.Digraph("graph", engine="neato", graph_attr=graph_attr)

    # colors of clusters
    ds = sorted([d for d in clusters.keys() if d is not None])
    if cmap is None:
        cluster_colors = {d: "black" for d in ds}
    else:
        cmap = plt.get_cmap(cmap)
        cs = np.linspace(0, 1, len(ds) + 1)
        cluster_colors = dict(zip(ds, cmap(cs)))
    cluster_colors[None] = cluster_colors[0]

    # distance -> y_position
    pos_y_map = {0: 0, None: 0}

    # distance from target to descendants -> pos_y
    dsc_dists = sorted([d for d in clusters.keys() if d is not None and 0 < d])
    for i, dist in enumerate(dsc_dists):
        pos_y_map[dist] = -(i + 1) * vmargin

    # distance from target to ascendants -> pos_y
    anc_dists = sorted([d for d in clusters.keys() if d is not None and d < 0], key=abs)
    for i, dist in enumerate(anc_dists):
        pos_y_map[dist] = (i + 1) * vmargin

    get_cluster_name = lambda dist: f"cluster{dist}"
    get_node_id = lambda cl, node: f"{cls_name}-{node}"

    node_ids = {node: [] for node in np.arange(adj.shape[0])}

    # add subgraphs and nodes
    for dist, nodes in clusters.items():
        # skip drawing the cluster
        if dist is None and draw_others is False:
            continue

        if len(nodes) == 0:
            continue

        # attributes for the cluster and nodes
        if dist == 0:
            cls_label = "target"
            offset_x = 0
        elif dist is None:
            cls_label = "others"
            offset_x = cls_margin_r + cls_margin_l + cls_space_target_other
            nodes = sorted(
                nodes, key=lambda x: (x in isolate_indices, x in other_indices, x)
            )
        else:
            cls_label = (
                f"ancestor {abs(dist)}" if dist < 0 else f"descendant {abs(dist)}"
            )
            offset_x = 0
            nodes = sorted(nodes)
        pos_y = pos_y_map[dist]
        color = mcolors.to_hex(cluster_colors[dist], keep_alpha=True)
        cls_name = get_cluster_name(dist)

        graph_attr = {
            "style": "dashed",
            "label": cls_label,
            "labeljust": "l",
            "color": color,
        }
        with graph.subgraph(name=cls_name, graph_attr=graph_attr) as c:
            # add nodes
            for i, node in enumerate(nodes):
                node_id = get_node_id(cls_name, node)
                pos_x = hmargin * i + offset_x
                kwargs = {}
                if path is not None:
                    if node in path_nodes:
                        kwargs["fontcolor"] = path_color
                        kwargs["color"] = path_color
                    if node == path_start or node == path_end:
                        kwargs["peripheries"] = "2"
                c.node(
                    node_id,
                    node_names[node],
                    pos=f"{pos_x},{pos_y}!",
                    color=color,
                    **kwargs,
                )
                node_ids[node].append((node_id, dist))

            # dummy at the upper left
            x = -cls_margin_l + offset_x
            y = pos_y + cls_margin_t
            c.node(
                f"{cls_name}-dummy_upper",
                "",
                pos=f"{x},{y}!",
                style="invis",
                fixedsize="true",
                width="0",
                height="0",
            )

            # dummy at the lower right
            x = (len(nodes) - 1) * hmargin + cls_margin_r + offset_x
            y = pos_y - cls_margin_b
            c.node(
                f"{cls_name}-dummy_lower",
                "",
                pos=f"{x},{y}!",
                style="invis",
                fixedsize="true",
                width="0",
                height="0",
            )

    # cycles
    cycle_edges = []
    if detect_cycle is True:
        for cycle in nx.simple_cycles(G):
            es = list(zip(cycle, np.roll(cycle, -1)))
            cycle_edges.append(es)
        cycle_edges = sum(cycle_edges, start=[])

    # indices of clusters
    other_nodes = isolate_indices | other_indices
    valid_nodes = other_nodes | dsc_indices | anc_indices | set([node_index])

    # add edges
    edges = np.argwhere(~np.isclose(adj, 0))
    for h, t in edges:
        if draw_others is False and (t in other_nodes or h in other_nodes):
            continue

        if t not in valid_nodes or h not in valid_nodes:
            continue

        for tail_id, dist_t in node_ids[t]:
            for head_id, dist_h in node_ids[h]:
                color = cluster_colors[dist_h]
                color = mcolors.to_hex(color, keep_alpha=True)

                kwargs = {"color": color}
                if path is not None and (t, h) in path_edges:
                    kwargs["penwidth"] = "2"
                    kwargs["fontcolor"] = path_color
                    kwargs["color"] = path_color
                if detect_cycle is True and (t, h) in cycle_edges:
                    kwargs["style"] = "dashed"

                graph.edge(tail_id, head_id, label=f"{adj[h, t]:.2f}", **kwargs)

    return graph


def predict_adaptive_lasso(X, predictors, target, gamma=1.0):
    """Predict with Adaptive Lasso.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples
        and n_features is the number of features.
    predictors : array-like, shape (n_predictors)
        Indices of predictor variable.
    target : int
        Index of target variable.

    Returns
    -------
    coef : array-like, shape (n_features)
        Coefficients of predictor variable.
    """
    # Standardize X
    scaler = StandardScaler()
    X_std = scaler.fit_transform(X)

    # Pruning with Adaptive Lasso
    lr = LinearRegression()
    lr.fit(X_std[:, predictors], X_std[:, target])
    weight = np.power(np.abs(lr.coef_), gamma)
    reg = LassoLarsIC(criterion="bic")
    reg.fit(X_std[:, predictors] * weight, X_std[:, target])
    pruned_idx = np.abs(reg.coef_ * weight) > 0.0

    # Calculate coefficients of the original scale
    coef = np.zeros(reg.coef_.shape)
    if pruned_idx.sum() > 0:
        lr = LinearRegression()
        pred = np.array(predictors)
        lr.fit(X[:, pred[pruned_idx]], X[:, target])
        coef[pruned_idx] = lr.coef_

    return coef


def find_all_paths(dag, from_index, to_index, min_causal_effect=0.0):
    """Find all paths from point to point in DAG.

    Parameters
    ----------
    dag : array-like, shape (n_features, n_features)
        The adjacency matrix to fine all paths, where n_features is the number of features.
    from_index : int
        Index of the variable at the start of the path.
    to_index : int
        Index of the variable at the end of the path.
    min_causal_effect : float, optional (default=0.0)
        Threshold for detecting causal direction.
        Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

    Returns
    -------
    paths : array-like, shape (n_paths)
        List of found path, where n_paths is the number of paths.
    effects : array-like, shape (n_paths)
        List of causal effect, where n_paths is the number of paths.
    """
    # Extract all edges
    edges = np.array(np.where(np.abs(np.nan_to_num(dag)) > min_causal_effect)).T

    # Aggregate edges by start point
    to_indices = []
    for i in range(dag.shape[0]):
        adj_list = edges[edges[:, 1] == i][:, 0].tolist()
        if len(adj_list) != 0:
            to_indices.append(adj_list)
        else:
            to_indices.append([])

    # DFS
    paths = []
    stack = [from_index]
    stack_to_indice = [to_indices[from_index]]
    while stack:
        if len(stack) > dag.shape[0]:
            raise ValueError(
                "Unable to find the path because a cyclic graph has been specified."
            )

        cur_index = stack[-1]
        to_indice = stack_to_indice[-1]

        if cur_index == to_index:
            paths.append(stack.copy())
            stack.pop()
            stack_to_indice.pop()
        else:
            if len(to_indice) > 0:
                next_index = to_indice.pop(0)
                stack.append(next_index)
                stack_to_indice.append(to_indices[next_index].copy())
            else:
                stack.pop()
                stack_to_indice.pop()

    # Calculate the causal effect for each path
    effects = []
    for p in paths:
        coefs = [dag[p[i + 1], p[i]] for i in range(len(p) - 1)]
        effects.append(np.cumprod(coefs)[-1])

    return paths, effects


def likelihood_i(x, i, b_i, bi_0):
    """Compute local log-likelihood of component i.

    Parameters
    ----------
    x : array-like, shape (n_features, n_samples)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    i : array-like
        Variable index.
    b_i : array-like
        The i^th column of adjacency matrix, B[i].
    bi_0 : float
        Constant value for the i^th variable.

    Return
    -------
    ll : float
        Local log-likelihood of component i.
    """

    sample_size = x.shape[1]  # number of data points
    var_i = variance_i(x, i, b_i)
    #
    ll = 0.0
    ll += np.sum(log_p_super_gaussian((x[i] - np.dot(b_i, x) - bi_0) / np.sqrt(var_i)))

    ll -= sample_size * np.log(np.sqrt(var_i))

    return ll


def log_p_super_gaussian(s):
    """Compute density function of the normalized independent components.

    Parameters
    ----------
    s : array-like, shape (1, n_samples)
        Data, where ``n_samples`` is the number of samples.

    Return
    -------
    x : float
        Density function of the normalized independent components,
        whose disturbances are super-Gaussian.
    """
    const = -0.35  # normalising constant
    return -np.sqrt(2.0) * np.absolute(s) + const


def variance_i(X, i, b_i):
    """Compute empirical variance of component i.

    Parameters
    ----------
    x : array-like, shape (n_features, n_samples)
        Data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    i : array-like
        Variable index.
    b_i : array-like
        The i^th column of adjacency matrix, B[i].

    Return
    -------
    variance : float
        Empirical variance of component i.
    """
    # T = X.shape[1]  # sample size
    estimated_disturbance = X[i] - np.dot(b_i, X)
    # variance = np.sum(estimated_disturbance ** 2) / T  # JMLR paper assumes zero mean
    variance = np.var(estimated_disturbance)  # stable version, even not zero mean.

    return variance


def evaluate_model_fit(adjacency_matrix, X, is_ordinal=None):
    """evaluate the given adjacency matrix and return fit indices

    Parameters
    ----------
    adjacency_matrix : array-like, shape (n_features, n_features)
        Adjacency matrix representing a causal graph.
        The i-th column and row correspond to the i-th column of X.
    X : array-like, shape (n_samples, n_features)
        Training data.
    is_ordinal : array-like, shape (n_features,)
        Binary list. The i-th element represents that the i-th column of X is ordinal or not.
        0 means not ordinal, otherwise ordinal.

    Return
    ------
    fit_indices : pandas.DataFrame
        Fit indices. This API uses semopy's calc_stats(). See semopy's reference for details.
    """

    # check inputs
    adj = check_array(adjacency_matrix, force_all_finite="allow-nan")
    if adj.ndim != 2 or (adj.shape[0] != adj.shape[1]):
        raise ValueError("adj must be an square matrix.")

    X = check_array(X)
    if X.shape[1] != adj.shape[1]:
        raise ValueError("X.shape[1] and adj.shape[1] must be the same.")

    if is_ordinal is None:
        is_ordinal = np.zeros(X.shape[1])
    else:
        is_ordinal = check_array(is_ordinal, ensure_2d=False).flatten()
    if is_ordinal.shape[0] != adj.shape[1]:
        raise ValueError("is_ordinal.shape[0] and adj.shape[1] must be the same.")

    # build desc
    desc = ""
    eta_names = []

    for i, row in enumerate(adj):
        # exogenous
        if np.sum(np.isnan(row)) == 0 and np.sum(np.isclose(row, 0)) == row.shape[0]:
            continue

        desc += f"x{i:d} ~ "

        for j, elem in enumerate(row):
            if np.isnan(elem):
                eta_name = f"eta_{i}_{j}" if i < j else f"eta_{j}_{i}"
                desc += f"{eta_name} + "
                if eta_name not in eta_names:
                    eta_names.append(eta_name)
            elif not np.isclose(elem, 0):
                desc += f"{elem:f} * x{j:d} + "
        desc = desc[: -len(" * ")] + "\n"

    if len(eta_names) > 0:
        desc += "DEFINE(latent) " + " ".join(eta_names) + "\n"

    if sum(is_ordinal) > 0:
        indices = np.argwhere(is_ordinal).flatten()

        desc += "DEFINE(ordinal)"
        for i in indices:
            desc += f" x{i}"
        desc += "\n"

    columns = [f"x{i:d}" for i in range(X.shape[1])]
    X = pd.DataFrame(X, columns=columns)

    m = semopy.Model(desc)
    m.fit(X)

    stats = semopy.calc_stats(m)

    return stats


def calculate_distance_from_root_nodes(adjacency_matrix, max_distance=None):
    """Calculate shortest distances from root nodes.

    Parameters
    ----------
    adjacency_matrices : array-like
        The adjacency matrix.
    max_distance : int or None, optional (default=None)
        The maximum distance to return nodes from root nodes.

    Returns
    -------
    shortest_distances : dict
        The dictionary has the following format::

        {'distance_from_root_node': [index_of_variables]}
    """
    # check inputs
    adjacency_matrix = check_array(
        adjacency_matrix,
        ensure_min_samples=2,
        ensure_min_features=2,
        force_all_finite="allow-nan",
    )
    if adjacency_matrix.shape[0] != adjacency_matrix.shape[1]:
        raise ValueError("adjacency_matrix must be an square matrix.")

    if max_distance is not None:
        max_distance = check_scalar(max_distance, "max_distance", int, min_val=1)

    # delete hidden common causes
    adjacency_matrix = np.nan_to_num(adjacency_matrix)

    # find root nodes
    root_indices = np.argwhere(
        np.isclose(np.sum(adjacency_matrix, axis=1), 0) == True
    ).flatten()
    if len(root_indices) == 0:
        raise ValueError("adjacency_matrix has no root nodes.")

    G = nx.from_numpy_array(adjacency_matrix.T, create_using=nx.DiGraph)

    # distances from root nodes
    dist_df = {}
    for index in root_indices:
        dist = nx.shortest_path_length(G, source=index)
        dist_df[index] = dist
    dist_df = pd.DataFrame(dist_df)

    dist_df = dist_df.fillna(np.iinfo(int).max)
    dist_df = dist_df.min(axis=1)
    if max_distance is not None:
        dist_df = dist_df.iloc[dist_df.values <= max_distance]
    dist_df = dist_df.sort_index()
    dist_df = dist_df.astype(int)

    result = {i: [] for i in pd.unique(dist_df)}
    for name, dist in dist_df.items():
        result[dist].append(name)

    return result


def calculate_total_effect(adjacency_matrix, from_index, to_index, is_continuous=None):
    """Calculate total effect.

    Parameters
    ----------
    adjacency_matrix : array_like
        The adjacency matrix.
    from_index : int
        The index of the cause variable.
    to_index : int
        The index of the effect variable.
    is_continuous : list
        The list of boolean. is_continuous indicates whether each variable
        is continuous or discrete.

    Returns
    -------
    total_effect : float
    """

    # check inputs
    adjacency_matrix = check_array(
        adjacency_matrix,
        ensure_min_samples=2,
        ensure_min_features=2,
        force_all_finite="allow-nan",
    )
    if adjacency_matrix.shape[0] != adjacency_matrix.shape[1]:
        raise ValueError(
            "adjacency_matrix must be an square matrix.", adjacency_matrix.shape
        )

    from_index = check_scalar(
        from_index,
        "from_index",
        (numbers.Integral, np.integer),
        min_val=0,
        max_val=len(adjacency_matrix) - 1,
    )

    to_index = check_scalar(
        to_index,
        "to_index",
        (numbers.Integral, np.integer),
        min_val=0,
        max_val=len(adjacency_matrix) - 1,
    )

    if from_index == to_index:
        raise ValueError("from_index and to_index mustn't be the same.")

    if is_continuous is None:
        is_continuous = [True for _ in range(len(adjacency_matrix))]
    else:
        is_continuous = check_array(
            is_continuous, ensure_2d=False, ensure_min_samples=len(adjacency_matrix)
        )

    # find all paths
    path_list, effects = find_all_paths(adjacency_matrix, from_index, to_index)

    # check all nodes on the path are continuous
    for path in path_list:
        for node in path[1:]:
            if not is_continuous[node]:
                raise ValueError("Variables on the path must be continuous variables.")

    total_effect = sum(effects)

    return total_effect

def get_cuda_version():
    try:
        nvcc_version = subprocess.check_output(["nvcc", "--version"]).decode('utf-8')
        print("CUDA Version found:\n", nvcc_version)
        return True
    except Exception as e:
        print("CUDA not found or nvcc not in PATH:", e)
        return False



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""
import itertools
import warnings

import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array, resample
from statsmodels.tsa.vector_ar.var_model import VAR

from .base import _BaseLiNGAM
from .bootstrap import BootstrapResult
from .direct_lingam import DirectLiNGAM
from .hsic import hsic_test_gamma
from .utils import predict_adaptive_lasso, find_all_paths, calculate_total_effect


class VARLiNGAM:
    """Implementation of VAR-LiNGAM Algorithm [1]_

    References
    ----------
    .. [1] Aapo Hyvärinen, Kun Zhang, Shohei Shimizu, Patrik O. Hoyer.
       Estimation of a Structural Vector Autoregression Model Using Non-Gaussianity.
       Journal of Machine Learning Research, 11: 1709-1731, 2010.
    """

    def __init__(
        self,
        lags=1,
        criterion="bic",
        prune=True,
        ar_coefs=None,
        lingam_model=None,
        random_state=None,
    ):
        """Construct a VARLiNGAM model.

        Parameters
        ----------
        lags : int, optional (default=1)
            Number of lags.
        criterion : {‘aic’, ‘fpe’, ‘hqic’, ‘bic’, None}, optional (default='bic')
            Criterion to decide the best lags within ``lags``.
            Searching the best lags is disabled if ``criterion`` is ``None``.
        prune : boolean, optional (default=True)
            Whether to prune the adjacency matrix of lags.
        ar_coefs : array-like, optional (default=None)
            Coefficients of AR model. Estimating AR model is skipped if specified ``ar_coefs``.
            Shape must be (``lags``, n_features, n_features).
        lingam_model : lingam object inherits 'lingam._BaseLiNGAM', optional (default=None)
            LiNGAM model for causal discovery. If None, DirectLiNGAM algorithm is selected.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        """
        self._lags = lags
        self._criterion = criterion
        self._prune = prune
        self._ar_coefs = (
            check_array(ar_coefs, allow_nd=True) if ar_coefs is not None else None
        )
        self._lingam_model = lingam_model
        self._random_state = random_state

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X: array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        returns
        -------
        self : object
            Returns the instance itself.
        """
        self._causal_order = None
        self._adjacency_matrices = None

        X = check_array(X)

        lingam_model = self._lingam_model
        if lingam_model is None:
            lingam_model = DirectLiNGAM()
        elif not isinstance(lingam_model, _BaseLiNGAM):
            raise ValueError("lingam_model must be a subclass of _BaseLiNGAM")

        M_taus = self._ar_coefs

        if M_taus is None:
            M_taus, lags, residuals = self._estimate_var_coefs(X)
        else:
            lags = M_taus.shape[0]
            residuals = self._calc_residuals(X, M_taus, lags)
            
        self._ar_coefs = M_taus
        self._lags = lags
        self._residuals = residuals

        model = lingam_model
        model.fit(residuals)

        B_taus = self._calc_b(X, model.adjacency_matrix_, M_taus)

        if self._prune:
            B_taus = self._pruning(X, B_taus, model.causal_order_)

        self._causal_order = model.causal_order_
        self._adjacency_matrices = B_taus

        return self

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : TimeseriesBootstrapResult
            Returns the result of bootstrapping.
        """
        X = check_array(X)

        n_samples = X.shape[0]
        n_features = X.shape[1]

        # store initial settings
        ar_coefs = self._ar_coefs
        lags = self._lags

        criterion = self._criterion
        self._criterion = None

        self.fit(X)

        fitted_ar_coefs = self._ar_coefs
        residuals = self._residuals
        
        total_effects = np.zeros(
            [n_sampling, n_features, n_features * (1 + self._lags)]
        )

        adjacency_matrices = []
        for i in range(n_sampling):
            sampled_residuals = resample(residuals, n_samples=n_samples)

            resampled_X = np.zeros((n_samples, n_features))
            for j in range(n_samples):
                if j < lags:
                    resampled_X[j, :] = sampled_residuals[j]
                    continue

                ar = np.zeros((1, n_features))
                for t, M in enumerate(fitted_ar_coefs):
                    ar += np.dot(M, resampled_X[j - t - 1, :].T).T

                resampled_X[j, :] = ar + sampled_residuals[j]

            # restore initial settings
            self._ar_coefs = ar_coefs
            self._lags = lags

            self.fit(resampled_X)
            am = np.concatenate([*self._adjacency_matrices], axis=1)
            adjacency_matrices.append(am)

            # total effects
            for c, to in enumerate(reversed(self._causal_order)):
                # time t
                for from_ in self._causal_order[: n_features - (c + 1)]:
                    total_effects[i, to, from_] = self.estimate_total_effect2(
                        n_features, from_, to
                    )

                # time t-tau
                for lag in range(self._lags):
                    for from_ in range(n_features):
                        total_effects[
                            i, to, from_ + n_features * (lag + 1)
                        ] = self.estimate_total_effect2(n_features, from_, to, lag + 1)

        self._criterion = criterion

        return VARBootstrapResult(adjacency_matrices, total_effects)

    def estimate_total_effect(self, X, from_index, to_index, from_lag=0):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        X = check_array(X)
        n_features = X.shape[1]

        # Check from/to causal order
        if from_lag == 0:
            from_order = self._causal_order.index(from_index)
            to_order = self._causal_order.index(to_index)
            if from_order > to_order:
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the causal order of the destination variable (to_index={to_index}) "
                    f"is earlier than the source variable (from_index={from_index})."
                )

        # X + lagged X
        X_joined = np.zeros((X.shape[0], X.shape[1] * (1 + self._lags + from_lag)))
        for p in range(1 + self._lags + from_lag):
            pos = n_features * p
            X_joined[:, pos : pos + n_features] = np.roll(X[:, 0:n_features], p, axis=0)

        # from_index + parents indices
        am = np.concatenate([*self._adjacency_matrices], axis=1)
        parents = np.where(np.abs(am[from_index]) > 0)[0]
        from_index = (
            from_index if from_lag == 0 else from_index + (n_features * from_lag)
        )
        parents = parents if from_lag == 0 else parents + (n_features * from_lag)
        predictors = [from_index]
        predictors.extend(parents)

        # estimate total effect
        lr = LinearRegression()
        lr.fit(X_joined[:, predictors], X_joined[:, to_index])

        return lr.coef_[0]

    def estimate_total_effect2(self, n_features, from_index, to_index, from_lag=0):
        """Estimate total effect using causal model.

        Parameters
        ----------
        n_features :
            The number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        # Check from/to causal order
        if from_lag == 0:
            from_order = self._causal_order.index(from_index)
            to_order = self._causal_order.index(to_index)
            if from_order > to_order:
                warnings.warn(
                    f"The estimated causal effect may be incorrect because "
                    f"the causal order of the destination variable (to_index={to_index}) "
                    f"is earlier than the source variable (from_index={from_index})."
                )

        # from_index + parents indices
        am = np.concatenate([*self._adjacency_matrices], axis=1)
        am = np.pad(am, [(0, am.shape[1] - am.shape[0]), (0, 0)])
        from_index = (
            from_index if from_lag == 0 else from_index + (n_features * from_lag)
        )

        effect = calculate_total_effect(am, from_index, to_index)

        return effect

    def get_error_independence_p_values(self):
        """Calculate the p-value matrix of independence between error variables.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        nn = self.residuals_
        B0 = self._adjacency_matrices[0]
        E = np.dot(np.eye(B0.shape[0]) - B0, nn.T).T
        n_samples = E.shape[0]
        n_features = E.shape[1]

        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            _, p_value = hsic_test_gamma(
                np.reshape(E[:, i], [n_samples, 1]), np.reshape(E[:, j], [n_samples, 1])
            )
            p_values[i, j] = p_value
            p_values[j, i] = p_value

        return p_values

    def _estimate_var_coefs(self, X):
        """Estimate coefficients of VAR"""
        # XXX: VAR.fit() is not searching lags correctly
        if self._criterion not in ["aic", "fpe", "hqic", "bic"]:
            var = VAR(X)
            result = var.fit(maxlags=self._lags, trend="n")
        else:
            min_value = float("Inf")
            result = None

            for lag in range(1, self._lags + 1):
                var = VAR(X)
                fitted = var.fit(maxlags=lag, ic=None, trend="n")

                value = getattr(fitted, self._criterion)
                if value < min_value:
                    min_value = value
                    result = fitted

        return result.coefs, result.k_ar, result.resid

    def _calc_residuals(self, X, M_taus, lags):
        """Calculate residulas"""
        X = X.T
        n_features = X.shape[0]
        n_samples = X.shape[1]

        residuals = np.zeros((n_features, n_samples))
        for t in range(n_samples):
            if t - lags < 0:
                continue

            estimated = np.zeros((X.shape[0], 1))
            for tau in range(1, lags + 1):
                estimated += np.dot(M_taus[tau - 1], X[:, t - tau].reshape((-1, 1)))

            residuals[:, t] = X[:, t] - estimated.reshape((-1,))

        residuals = residuals[:, lags:].T

        return residuals

    def _calc_b(self, X, B0, M_taus):
        """Calculate B"""
        n_features = X.shape[1]

        B_taus = np.array([B0])

        for M in M_taus:
            B_t = np.dot((np.eye(n_features) - B0), M)
            B_taus = np.append(B_taus, [B_t], axis=0)

        return B_taus

    def _pruning(self, X, B_taus, causal_order):
        """Prune edges"""
        n_features = X.shape[1]

        stacked = [np.flip(X, axis=0)]
        for i in range(self._lags):
            stacked.append(np.roll(stacked[-1], -1, axis=0))
        blocks = np.array(list(zip(*stacked)))[: -self._lags]

        for i in range(n_features):
            causal_order_no = causal_order.index(i)
            ancestor_indexes = causal_order[:causal_order_no]

            obj = np.zeros((len(blocks)))
            exp = np.zeros((len(blocks), causal_order_no + n_features * self._lags))
            for j, block in enumerate(blocks):
                obj[j] = block[0][i]
                exp[j:] = np.concatenate(
                    [block[0][ancestor_indexes].flatten(), block[1:][:].flatten()],
                    axis=0,
                )

            # adaptive lasso
            predictors = [i for i in range(exp.shape[1])]
            target = len(predictors)
            X_con = np.concatenate([exp, obj.reshape(-1, 1)], axis=1)
            coef = predict_adaptive_lasso(X_con, predictors, target)

            B_taus[0][i, ancestor_indexes] = coef[:causal_order_no]
            for j in range(len(B_taus[1:])):
                B_taus[j + 1][i, :] = coef[
                    causal_order_no + n_features * j :
                    causal_order_no + n_features * j + n_features
                ]

        return B_taus

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
        """
        return self._causal_order

    @property
    def adjacency_matrices_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (lags, n_features, n_features)
            The adjacency matrix of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrices

    @property
    def residuals_(self):
        """Residuals of regression.

        Returns
        -------
        residuals_ : array-like, shape (n_samples)
            Residuals of regression, where n_samples is the number of samples.
        """
        return self._residuals


class VARBootstrapResult(BootstrapResult):
    """The result of bootstrapping for Time series algorithm."""

    def __init__(self, adjacency_matrices, total_effects):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        """
        super().__init__(adjacency_matrices, total_effects)

    def get_paths(
        self, from_index, to_index, from_lag=0, to_lag=0, min_causal_effect=None
    ):
        """Get all paths from the start variable to the end variable and their bootstrap probabilities.

        Parameters
        ----------
        from_index : int
            Index of the variable at the start of the path.
        to_index : int
            Index of the variable at the end of the path.
        from_lag : int
            Number of lag at the start of the path.
            ``from_lag`` should be greater than or equal to ``to_lag``.
        to_lag : int
            Number of lag at the end of the path.
            ``from_lag`` should be greater than or equal to ``to_lag``.
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        paths : dict
            List of path and bootstrap probability.
            The dictionary has the following format::

            {'path': [n_paths], 'effect': [n_paths], 'probability': [n_paths]}

            where ``n_paths`` is the number of paths.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")
        if to_lag > from_lag:
            raise ValueError("from_lag should be greater than or equal to to_lag.")
        if to_lag == from_lag:
            if to_index == from_index:
                raise ValueError("The same variable is specified for from and to.")

        # Find all paths from from_index to to_index
        paths_list = []
        effects_list = []
        for am in self._adjacency_matrices:
            expansion_m = np.zeros((am.shape[1], am.shape[1]))
            n_features = am.shape[0]
            n_lags = int(am.shape[1] / am.shape[0]) - 1
            for i in range(n_lags + 1):
                for j in range(i, n_lags + 1):
                    row = n_features * i
                    col = n_features * j
                    lag = col - row
                    expansion_m[row : row + n_features, col : col + n_features] = am[
                        0:n_features, lag : lag + n_features
                    ]
            paths, effects = find_all_paths(
                expansion_m,
                int(n_features * from_lag + from_index),
                int(n_features * to_lag + to_index),
                min_causal_effect,
            )

            # Convert path to string to make them easier to handle.
            paths_list.extend(["_".join(map(str, p)) for p in paths])
            effects_list.extend(effects)

        paths_list = np.array(paths_list)
        effects_list = np.array(effects_list)

        # Count paths
        paths_str, counts = np.unique(paths_list, axis=0, return_counts=True)

        # Sort by count
        order = np.argsort(-counts)
        probs = counts[order] / len(self._adjacency_matrices)
        paths_str = paths_str[order]

        # Calculate median of causal effect for each path
        effects = [
            np.median(effects_list[np.where(paths_list == p)]) for p in paths_str
        ]

        result = {
            "path": [[int(i) for i in p.split("_")] for p in paths_str],
            "effect": effects,
            "probability": probs.tolist(),
        }
        return result



import os

import numpy as np
import pandas as pd
from lingam.group_direct_lingam import GroupDirectLiNGAM


def test_fit_success():
    # causal direction: x0, x1 --> x2 --> x3, x4
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = 0.5 * x1 + np.random.uniform(-1, 1, size=1000)
    x3 = 0.7 * x2 + np.random.uniform(-2, 2, size=1000)
    x4 = 1.5 * x0 + 0.8 * x3 + np.random.uniform(-2, 2, size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])
    groups = [[0, 1], [2], [3, 4]]

    model = GroupDirectLiNGAM()
    model.fit(X, groups)

    co = model.causal_order_
    am = model.adjacency_matrix_

def test_fit_invalid():
    # Not array data
    X = 1
    groups = [[0, 1], [2], [3, 4]]
    try:
        model = GroupDirectLiNGAM()
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    groups = [[0], [1]]
    try:
        model = GroupDirectLiNGAM()
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = GroupDirectLiNGAM()
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = GroupDirectLiNGAM()
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    # causal direction: x0, x1 --> x2 --> x3, x4
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = 0.5 * x1 + np.random.uniform(-1, 1, size=1000)
    x3 = 0.7 * x2 + np.random.uniform(-2, 2, size=1000)
    x4 = 1.5 * x0 + 0.8 * x3 + np.random.uniform(-2, 2, size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])
    groups = [[0, 1], [2], [3, 4]]

    model = GroupDirectLiNGAM()
    result = model.bootstrap(X, groups, n_sampling=5)
    am = result.adjacency_matrices_
    assert len(am) == 5
    te = result.total_effects_
    assert len(te) == 5

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    paths = result.get_paths(1, 2)

def test_bootstrap_invalid_data():
    # causal direction: x0, x1 --> x2 --> x3, x4
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = 0.5 * x1 + np.random.uniform(-1, 1, size=1000)
    x3 = 0.7 * x2 + np.random.uniform(-2, 2, size=1000)
    x4 = 1.5 * x0 + 0.8 * x3 + np.random.uniform(-2, 2, size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])
    groups = [[0, 1], [2], [3, 4]]

    # Invalid argument: bootstrap(n_sampling=-1)
    model = GroupDirectLiNGAM()
    try:
        result = model.bootstrap(X, groups, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = GroupDirectLiNGAM()
    try:
        result = model.bootstrap(X, groups, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = GroupDirectLiNGAM()
    result = model.bootstrap(X, groups, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    try:
        result.get_paths(0, 1, min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_prior_knowledge_success():
    # causal direction: x0, x1 --> x2 --> x3, x4
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = 0.5 * x1 + np.random.uniform(-1, 1, size=1000)
    x3 = 0.7 * x2 + np.random.uniform(-2, 2, size=1000)
    x4 = 1.5 * x0 + 0.8 * x3 + np.random.uniform(-2, 2, size=1000)
    x5 = -0.6 * x3 - 0.5 * x4 + np.random.uniform(-3, 3, size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4, x5]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4', 'x5'])
    groups = [[0, 1], [2], [3, 4], [5]]

    # prior knowledge: nothing
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    model = GroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X, groups)

    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # prior knowledge: x1 is exogenous
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, 0, 0],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    model = GroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X, groups)
    co = model.causal_order_
    assert co.index(1) < co.index(0)
    assert co.index(1) < co.index(2)
    assert co.index(1) < co.index(3)

    # prior knowledge: x0 is sink
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, -1, -1],
            [0, -1, 0, -1],
            [0, -1, -1, 0],
        ]
    )

    model = GroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X, groups)
    co = model.causal_order_
    assert co.index(0) > co.index(1)
    assert co.index(0) > co.index(2)
    assert co.index(0) > co.index(3)

    # prior knowledge: x2-->x3 has path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, 0],
            [-1, -1, 1, 0],
        ]
    )

    model = GroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X, groups)
    co = model.causal_order_
    assert co.index(2) < co.index(3)

    # prior knowledge: x1-->x3 does not have path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, 0, -1, 0],
        ]
    )

    model = GroupDirectLiNGAM(prior_knowledge=pk)
    model.fit(X, groups)

def test_prior_knowledge_invalid():
    # causal direction: x0, x1 --> x2 --> x3, x4
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = 0.5 * x1 + np.random.uniform(-1, 1, size=1000)
    x3 = 0.7 * x2 + np.random.uniform(-2, 2, size=1000)
    x4 = 1.5 * x0 + 0.8 * x3 + np.random.uniform(-2, 2, size=1000)
    x5 = -0.6 * x3 - 0.5 * x4 + np.random.uniform(-3, 3, size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4, x5]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4', 'x5'])
    groups = [[0, 1], [2], [3, 4], [5]]

    # prior knowledge: invalid
    pk = np.array(
        [
            [0, -1, 1],
            [-1, 0, -1],
            [-1, -1, 0],
            [-1, -1, -1],
        ]
    )

    try:
        model = GroupDirectLiNGAM(prior_knowledge=pk)
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError

    # prior knowledge: inconsistent
    pk = np.array(
        [
            [0, 1, -1, -1],
            [1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )

    try:
        model = GroupDirectLiNGAM(prior_knowledge=pk)
        model.fit(X, groups)
    except ValueError:
        pass
    else:
        raise AssertionError



import os

import numpy as np
import pandas as pd
from lingam.resit import RESIT
from sklearn.ensemble import RandomForestRegressor


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = RESIT(regressor=reg)
    model.fit(X)

    # check the causal ordering
    co = model.causal_order_

    # check the adjacency matrix
    am = model.adjacency_matrix_

    # check estimate_total_effect (Not implement)
    te = model.estimate_total_effect(X, 0, 3)

    # check get_error_independence_p_values (Not implement)
    p_values = model.get_error_independence_p_values(X)


def test_fit_invalid():
    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # Not array data
    X = 1
    try:
        model = RESIT(regressor=reg)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    # Invalid regressor
    try:
        model = RESIT(regressor=None)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid regressor
    dummy_reg = lambda x: x
    try:
        model = RESIT(regressor=dummy_reg)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid alpha
    try:
        model = RESIT(regressor=reg, alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_bootstrap_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)
    model = RESIT(regressor=reg)
    result = model.bootstrap(X, n_sampling=5)
    am = result.adjacency_matrices_
    assert len(am) == 5
    te = result.total_effects_
    assert len(te) == 5

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    paths = result.get_paths(1, 2)


def test_bootstrap_invalid_data():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # Invalid argument: bootstrap(n_sampling=-1)
    model = RESIT(regressor=reg)
    try:
        result = model.bootstrap(X, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = RESIT(regressor=reg)
    try:
        result = model.bootstrap(X, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = RESIT(regressor=reg)
    result = model.bootstrap(X, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    try:
        result.get_paths(0, 1, min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_prior_knowledge_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # prior knowledge: nothing
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    model = RESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_

    # prior knowledge: x1 is exogenous
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, 0, 0],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    model = RESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    assert co.index(1) < co.index(0)
    assert co.index(1) < co.index(2)
    assert co.index(1) < co.index(3)

    # prior knowledge: x0 is sink
    pk = np.array(
        [
            [0, -1, -1, -1],
            [0, 0, -1, -1],
            [0, -1, 0, -1],
            [0, -1, -1, 0],
        ]
    )
    model = RESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    assert co.index(0) > co.index(1)
    assert co.index(0) > co.index(2)
    assert co.index(0) > co.index(3)

    # prior knowledge: x2-->x3 has path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, 0],
            [-1, -1, 1, 0],
        ]
    )
    model = RESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_
    assert co.index(2) < co.index(3)

    # prior knowledge: x1-->x3 does not have path
    pk = np.array(
        [
            [0, -1, -1, -1],
            [-1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, 0, -1, 0],
        ]
    )
    model = RESIT(regressor=reg, prior_knowledge=pk)
    model.fit(X)
    co = model.causal_order_


def test_prior_knowledge_invalid():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=["x0", "x1", "x2", "x3"])

    reg = RandomForestRegressor(max_depth=4, random_state=0)

    # prior knowledge: invalid
    pk = np.array(
        [
            [0, -1, 1],
            [-1, 0, -1],
            [-1, -1, 0],
            [-1, -1, -1],
        ]
    )
    try:
        model = RESIT(regressor=reg, prior_knowledge=pk)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # prior knowledge: inconsistent
    pk = np.array(
        [
            [0, 1, -1, -1],
            [1, 0, -1, -1],
            [-1, -1, 0, -1],
            [-1, -1, -1, 0],
        ]
    )
    try:
        model = RESIT(regressor=reg, prior_knowledge=pk)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError



import os

import numpy as np
import random
import lingam


def get_noise(n):
    noise = ((np.random.rand(1, n) - 0.5) * 5).reshape(n)
    mean = get_random_constant(0.0, 2.0)
    noise += mean
    return noise

def causal_func(cause):
    a = get_random_constant(-5.0, 5.0)
    b = get_random_constant(-1.0, 1.0)
    c = int(random.uniform(2, 3))
    return ((cause + a) ** (c)) + b

def get_random_constant(s, b):
    constant = random.uniform(-1.0, 1.0)
    if constant > 0:
        constant = random.uniform(s, b)
    else:
        constant = random.uniform(-b, -s)
    return constant

def create_data(n):
    causal_pairs = [[0, 1], [0, 3], [2, 4]]
    intermediate_pairs = [[2, 5]]
    confounder_pairs = [[3, 4]]

    n_variables = 6

    data = np.zeros((n, n_variables))  # observed data
    confounders = np.zeros(
        (n, len(confounder_pairs))
    )  # data of unobserced common causes

    # Adding external effects
    for i in range(n_variables):
        data[:, i] = get_noise(n)
    for i in range(len(confounder_pairs)):
        confounders[:, i] = get_noise(n)
        confounders[:, i] = confounders[:, i] / np.std(confounders[:, i])

    # Adding the effects of unobserved common causes
    for i, cpair in enumerate(confounder_pairs):
        cpair = list(cpair)
        cpair.sort()
        data[:, cpair[0]] += causal_func(confounders[:, i])
        data[:, cpair[1]] += causal_func(confounders[:, i])

    for i1 in range(n_variables)[0:n_variables]:
        data[:, i1] = data[:, i1] / np.std(data[:, i1])
        for i2 in range(n_variables)[i1 + 1 : n_variables + 1]:
            # Adding direct effects between observed variables
            if [i1, i2] in causal_pairs:
                data[:, i2] += causal_func(data[:, i1])
            # Adding undirected effects between observed variables mediated through unobserved variables
            if [i1, i2] in intermediate_pairs:
                interm = causal_func(data[:, i1]) + get_noise(n)
                interm = interm / np.std(interm)
                data[:, i2] += causal_func(interm)

    return data

def test_fit_success():
    X = create_data(200)
    model = lingam.CAMUV()
    model.fit(X)
    print(model.adjacency_matrix_)

    # f-correlation
    model = lingam.CAMUV(independence="fcorr", ind_corr=0.5)
    model.fit(X)

    # prior_knowledge
    model = lingam.CAMUV(prior_knowledge=[(1, 0)])
    model.fit(X)

def test_fit_invalid():
    try:
        X = create_data(200)
        model = lingam.CAMUV(alpha=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        X = create_data(200)
        model = lingam.CAMUV(num_explanatory_vals=-1)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: independence
    try:
        model = lingam.CAMUV(independence="lingam")
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    try:
        X = create_data(200)
        model = lingam.CAMUV(ind_corr=-1.0)
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numpy as np
from scipy.optimize import linear_sum_assignment
from sklearn.utils import check_array
from sklearn.decomposition import FastICA

from .base import _BaseLiNGAM


class ICALiNGAM(_BaseLiNGAM):
    """Implementation of ICA-based LiNGAM Algorithm [1]_

    References
    ----------
    .. [1] S. Shimizu, P. O. Hoyer, A. Hyvärinen, and A. J. Kerminen.
       A linear non-gaussian acyclic model for causal discovery.
       Journal of Machine Learning Research, 7:2003-2030, 2006.
    """

    def __init__(self, random_state=None, max_iter=1000):
        """Construct a ICA-based LiNGAM model.

        Parameters
        ----------
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        max_iter : int, optional (default=1000)
            The maximum number of iterations of FastICA.
        """
        super().__init__(random_state)
        self._max_iter = max_iter

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance of self.
        """
        X = check_array(X)

        # obtain a unmixing matrix from the given data
        ica = FastICA(max_iter=self._max_iter, random_state=self._random_state)
        ica.fit(X)
        W_ica = ica.components_

        # obtain a permuted W_ica
        _, col_index = linear_sum_assignment(1 / np.abs(W_ica))
        PW_ica = np.zeros_like(W_ica)
        PW_ica[col_index] = W_ica

        # obtain a vector to scale
        D = np.diag(PW_ica)[:, np.newaxis]

        # estimate an adjacency matrix
        W_estimate = PW_ica / D
        B_estimate = np.eye(len(W_estimate)) - W_estimate

        causal_order = self._estimate_causal_order(B_estimate)
        self._causal_order = causal_order

        return self._estimate_adjacency_matrix(X)

    def _search_causal_order(self, matrix):
        """Obtain a causal order from the given matrix strictly.

        Parameters
        ----------
        matrix : array-like, shape (n_features, n_samples)
            Target matrix.

        Return
        ------
        causal_order : array, shape [n_features, ]
            A causal order of the given matrix on success, None otherwise.
        """
        causal_order = []

        row_num = matrix.shape[0]
        original_index = np.arange(row_num)

        while 0 < len(matrix):
            # find a row all of which elements are zero
            row_index_list = np.where(np.sum(np.abs(matrix), axis=1) == 0)[0]
            if len(row_index_list) == 0:
                break

            target_index = row_index_list[0]

            # append i to the end of the list
            causal_order.append(original_index[target_index])
            original_index = np.delete(original_index, target_index, axis=0)

            # remove the i-th row and the i-th column from matrix
            mask = np.delete(np.arange(len(matrix)), target_index, axis=0)
            matrix = matrix[mask][:, mask]

        if len(causal_order) != row_num:
            causal_order = None

        return causal_order

    def _estimate_causal_order(self, matrix):
        """Obtain a lower triangular from the given matrix approximately.

        Parameters
        ----------
        matrix : array-like, shape (n_features, n_samples)
            Target matrix.

        Return
        ------
        causal_order : array, shape [n_features, ]
            A causal order of the given matrix on success, None otherwise.
        """
        causal_order = None

        # set the m(m + 1)/2 smallest(in absolute value) elements of the matrix to zero
        pos_list = np.argsort(np.abs(matrix), axis=None)
        pos_list = np.vstack(np.unravel_index(pos_list, matrix.shape)).T
        initial_zero_num = int(matrix.shape[0] * (matrix.shape[0] + 1) / 2)
        for i, j in pos_list[:initial_zero_num]:
            matrix[i, j] = 0

        for i, j in pos_list[initial_zero_num:]:
            causal_order = self._search_causal_order(matrix)
            if causal_order is not None:
                break
            else:
                # set the smallest(in absolute value) element to zero
                matrix[i, j] = 0

        return causal_order



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numbers

import numpy as np
from scipy.stats import laplace
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array, resample

from .utils import predict_adaptive_lasso, calculate_total_effect
from .bootstrap import BootstrapResult


class GroupDirectLiNGAM:
    """Implementation of GroupDirectLiNGAM Algorithm [1]_

    References
    ----------
    .. [1] D. Entner and P. O. Hoyer. Estimating a causal order among groups of variables in linear models.
       In Proc. 22nd International Conference on Artificial Neural Networks (ICANN2012),
       pp. 83--90, Lausanne, Switzerland, 2012.
    """
    def __init__(self, prior_knowledge=None):
        """Construct a GroupDirectLiNGAM model.

        Parameters
        ----------
        prior_knowledge : array-like, shape (n_groups, n_groups), optional (default=None)
            Prior knowledge used for causal discovery, where ``n_groups`` is the number of groups.

            The elements of prior knowledge matrix are defined as follows [1]_:

            * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
            * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
            * ``-1`` : No prior knowledge is available to know if either of the two cases above (0 or 1) is true.
        """
        self._adjacency_matrix = None
        self._causal_order = []

        self._Aknw = prior_knowledge
        if self._Aknw is not None:
            self._Aknw = check_array(self._Aknw)
            self._Aknw = np.where(self._Aknw < 0, np.nan, self._Aknw)

    def fit(self, X, groups):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        groups : array-like, shape (n_groups)
            The list of features for each group.
            where ``n_groups`` is the number of groups.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)
        n_groups = len(groups)

        # Check prior knowledge
        if self._Aknw is not None:
            if (n_groups, n_groups) != self._Aknw.shape:
                raise ValueError(
                    "The shape of prior knowledge must be (n_groups, n_groups)"
                )
            else:
                # Extract all partial orders in prior knowledge matrix
                self._partial_orders = self._extract_partial_orders(self._Aknw)

        # Causal discovery
        self._causal_order = self._estimate_causal_order(X, groups)

        return self._estimate_adjacency_matrix(X, groups, prior_knowledge=self._Aknw)

    def _extract_partial_orders(self, pk):
        """Extract partial orders from prior knowledge."""
        path_pairs = np.array(np.where(pk == 1)).transpose()
        no_path_pairs = np.array(np.where(pk == 0)).transpose()

        # Check for inconsistencies in pairs with path
        check_pairs = np.concatenate([path_pairs, path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            if len(pairs[counts > 1]) > 0:
                raise ValueError(
                    f"The prior knowledge contains inconsistencies at the following indices: {pairs[counts>1].tolist()}"
                )

        # Check for inconsistencies in pairs without path
        # If there are duplicate pairs without path, they cancel out and are not ordered.
        check_pairs = np.concatenate([no_path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            check_pairs = np.concatenate([no_path_pairs, pairs[counts > 1]])
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            no_path_pairs = pairs[counts < 2]

        check_pairs = np.concatenate([path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) == 0:
            # If no pairs are extracted from the specified prior knowledge,
            return check_pairs

        pairs = np.unique(check_pairs, axis=0)
        return pairs[:, [1, 0]]  # [to, from] -> [from, to]

    def _search_candidate(self, S):
        """Search for candidate features"""
        # If no prior knowledge is specified, nothing to do.
        if self._Aknw is None:
            return S

        # Candidate features that are not to the left of the partial orders
        if len(self._partial_orders) != 0:
            Sc = [i for i in S if i not in self._partial_orders[:, 1]]
            return Sc

        return S

    def _get_resid_and_coef(self, predictors, target):
        """Get the residuals and coefficients of the ordinary least squares method"""
        lr = LinearRegression().fit(predictors, target)
        resid = target - lr.predict(predictors)
        return resid, lr.coef_

    def _log_likelihood(self, x, y):
        """Compute the log likelihoods assuming a Laplace distribution"""
        residuals, _ = self._get_resid_and_coef(y, x)
        sigma = np.mean(np.abs(residuals))
        return np.sum(laplace.logpdf(residuals, loc=0, scale=sigma))

    def _log_likelihood_ratio(self, x, y):
        """Compute the (normalized) ratio of the log likelihoods for the two causal models"""
        if x.ndim == 1:
            x = x.reshape(-1, 1)
        if y.ndim == 1:
            y = y.reshape(-1, 1)

        m = len(x)
        LogL_xy = self._log_likelihood(x, y)
        LogL_yx = self._log_likelihood(y, x)

        R = (LogL_xy - LogL_yx) / m
        return R

    def _pairwise_measure(self, X, remaining_groups, j, groups):
        """Compute the pairwise measure between one group and all remaining groups"""
        mu = 0
        total_n_i = 0
        X_j = X[:, groups[j]]
        n_j = X_j.shape[1]
        for i in remaining_groups:
            if i == j:
                continue
            X_i = X[:, groups[i]]
            n_i = X_i.shape[1]
            total_n_i += n_i
            for k in range(n_j):
                x_jk = X_j[:, k]
                for l in range(n_i):
                    x_il = X_i[:, l]
                    r_li, coef = self._get_resid_and_coef(X_j, x_il)
                    b_lk = coef[k]
                    z_ikl = b_lk * x_jk + r_li
                    R = self._log_likelihood_ratio(x_jk, z_ikl)
                    mu += min(0, R) ** 2
        return mu / (n_j * total_n_i)

    def _get_residual_matrix(self, X, predictors, targets):
        """Get the matrix of residuals"""
        if len(predictors) == 0:
            return X

        Y = np.zeros_like(X)
        for target in targets:
            Y[:, target], _ = self._get_resid_and_coef(X[:, list(predictors)], X[:, target])
        return Y

    def _estimate_causal_order(self, X, groups):
        """Estimate the causal order of groups by repeatedly identifying exogenous groups"""
        n_groups = len(groups)
        G = np.arange(n_groups)
        Rj = np.copy(X)
        S = G
        K = []
        for _ in range(n_groups):
            Sc = self._search_candidate(S)

            if len(Sc) == 1:
                k = Sc[0]
            else:
                # Find an exogenous group
                mu_list = [float('inf') for j in range(n_groups)]
                for j in G:
                    if j in Sc:
                        mu = self._pairwise_measure(Rj, Sc, j, groups)
                        mu_list[j] = mu

                k = np.argmin(mu_list)

            S = S[S != k]
            K.append(k)

            # Replace the data matrix with the residuals
            Xj = groups[k]
            Xi = [k for i in S for k in groups[i]]
            Rj = self._get_residual_matrix(Rj, Xj, Xi)

            # Update partial orders
            if self._Aknw is not None:
                self._partial_orders = self._partial_orders[self._partial_orders[:, 0] != k]

        return K

    def _estimate_adjacency_matrix(self, X, groups, prior_knowledge=None):
        """Estimate adjacency matrix by causal order.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        groups : array-like, shape (n_groups)
            The list of features for each group.
        prior_knowledge : array-like, shape (n_groups, n_groups), optional (default=None)
            Prior knowledge matrix.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        if prior_knowledge is not None:
            pk = prior_knowledge.copy()
            np.fill_diagonal(pk, 0)

        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i in range(1, len(self._causal_order)):
            target_group_idx = self._causal_order[i]
            predictor_groups_idxs = self._causal_order[:i]

            # Exclude groups specified in no_path with prior knowledge
            if prior_knowledge is not None:
                predictor_groups_idxs = [p for p in predictor_groups_idxs if pk[target_group_idx, p] != 0]

            # target is exogenous groups if predictors are empty
            if len(predictor_groups_idxs) == 0:
                continue

            # Retrieve variables from groups
            target_group = groups[target_group_idx]
            predictor_groups = [groups[idx] for idx in predictor_groups_idxs]
            predictors = [var for group in predictor_groups for var in group]

            for target in target_group:
                B[target, predictors] = predict_adaptive_lasso(X, predictors, target)

        self._adjacency_matrix = B
        return self

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrix

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_groups)
            The causal order of fitted model, where
            n_groups is the number of groups.
        """
        return self._causal_order

    def bootstrap(self, X, groups, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        groups : array-like, shape (n_groups)
            The list of features for each group.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        index = np.arange(X.shape[0])
        resampled_indices = []
        for i in range(n_sampling):
            resampled_X, resampled_index = resample(X, index)
            self.fit(resampled_X, groups)
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for c, from_group in enumerate(self._causal_order):
                for to_group in self._causal_order[c + 1 :]:
                    for from_ in groups[from_group]:
                        for to in groups[to_group]:
                            total_effects[i, to, from_] = calculate_total_effect(
                                self._adjacency_matrix, from_, to
                            )

            resampled_indices.append(resampled_index)

        return BootstrapResult(adjacency_matrices, total_effects, resampled_indices=resampled_indices)



from .direct_lingam import DirectLiNGAM
from .ica_lingam import ICALiNGAM
from .bottom_up_parce_lingam import BottomUpParceLiNGAM
from .var_lingam import VARLiNGAM

from abc import ABCMeta, abstractmethod

import numpy as np
import pandas as pd
from scipy.special import expit

from sklearn.utils import check_array, check_random_state
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.base import clone, is_regressor, is_classifier
from sklearn.pipeline import Pipeline
from sklearn.model_selection._search import BaseSearchCV


class CausalBasedSimulator:
    """
    Causal based simulator.

    Attributes
    ----------
    train_result_ : dict
        Information about trained models.
    """

    def train(self, X, causal_graph, cd_algo_name=None, models=None, is_discrete=None):
        """
        Estimate functional relations between variables and variable
        distributions based on the training data ``X`` and the causal graph.
        The functional relations represents by linear regression model
        if the object variable is numeric, and represents by logistic
        regression model if the object variable is discrete by default.
        ``train_result_`` will be exposed after executing train().

        Parameters
        ----------
        X : array-like
            Training data.

        causal_graph : array-like
            Causal graph.

        cd_algo_name : str
            Algorithm type.

        models : dict, default=None
            Dictionary about the model of a variable in X. Models are cloned internaly
            and are trained to infer functioal relations. Given instances of
            the model are cloned to estimate the functional relation between
            variables.

        Returns
        -------
        self : Object
        """

        if cd_algo_name is None:
            cd_algo_name = "DirectLiNGAM"
        if not isinstance(cd_algo_name, str):
            raise TypeError("cd_algo_name must be str or None.")

        impl_constructor = self._dispatch_impl(cd_algo_name)
        impl = impl_constructor(X, causal_graph, is_discrete=is_discrete)

        train_models = self._check_models(models, impl.endog_names_, impl.discrete_endog_names_)

        train_result = self._train(train_models, causal_graph, impl)

        self._impl = impl
        self._train_result = train_result
        return self

    def run(
        self,
        changing_exog=None,
        changing_models=None,
        shuffle_residual=False,
        random_state=None,
    ):
        """
        Generate simulated data using trained models and the given
        causal graph with given exogenous data and models.
        Specifying environmental changes to ``changing_exog`` or
        specifiyig changes in fucitonal relation to ``change_models``
        effects simulated data.
        Residuals to simulate variables are shuffled using radom_state
        if ``shuffle_residual`` is True.

        Parameters
        ----------
        changing_exog : dict, default=None
            Dictioary about exogeous variables which keys are variable
            names and values are data of variables. That variable name
            should be a one of column names of X and the length should
            be same as X.

        changing_model : dict, default=None
            Dictionary about the changing models which elements are dictionary.
            The key is the name of the target variable and that value is
            a dictionary for that model. This dictionary contains three keys,
            parant_names, coef and model. parent_names is the mandatory key and
            whose value is the list of parent names. coef and model are selective.
            coef is the list of the coefficients of the parent variables, which
            must be same lenght as parent_names. The value of model must be
            a trained machine laerning instance.

        shuffle_residual : bool, default=True
            If True, residuals are shuffled.

        random_state : int, default=None
            If shuffle_residual is True, random_state is used as seed.

        Returns
        -------
        simulated_data : pandas.DataFrame
            simulated data.
        """

        if self._train_result is None:
            raise RuntimeError("run() must be executed after train() is executed")

        random_state = check_random_state(random_state)

        # check inputs
        changing_exog = self._check_changing_exog(
            changing_exog,
            self._impl.exog_length_,
            self._impl.endog_names_,
            self._impl.discrete_endog_names_
        )

        changing_models = self._check_changing_models(
            changing_models,
            self._impl.endog_names_,
            self._impl.discrete_endog_names_
        )

        simulated_data = self._simulate(
            changing_exog,
            changing_models,
            self._impl,
            self._train_result,
            shuffle_residual,
            random_state,
        )

        return simulated_data

    @property
    def train_result_(self):
        return self._train_result

    def _check_model_instance(self, model, var_name, discrete_endog_names):
        if var_name not in discrete_endog_names:
            check_model_type = is_regressor
        else:
            check_model_type = is_classifier

        if isinstance(model, Pipeline):
            if not check_model_type(model.steps[-1][-1]):
                mes = "The last step in Pipeline should be an "
                mes += "instance of a regression/classification model."
                raise RuntimeError(mes)
        elif isinstance(model, BaseSearchCV):
            if not check_model_type(model.get_params()["estimator"]):
                mes = "The type of the estimator shall be an "
                mes += "instance of a regression/classification model."
                raise RuntimeError(mes)
        else:
            if not check_model_type(model):
                mes = "The type of the estimator shall be an "
                mes += "instance of a regression/classification model."
                raise RuntimeError(mes)

        if check_model_type == is_classifier:
            try:
                func = getattr(model, "predict_proba")
                if not callable(func):
                    raise Exception
            except Exception:
                mes = "Classification models shall have "
                mes += "predict_proba()."
                raise RuntimeError(mes)

    def _check_models(self, models, endog_names, discrete_endog_names):
        if models is None:
            return {}

        if not isinstance(models, dict):
            raise RuntimeError("models must be a dictionary.")

        for var_name, model in models.items():
            if var_name not in endog_names:
                raise RuntimeError(f"Unknown variable name ({var_name})")

            self._check_model_instance(model, var_name, discrete_endog_names)

        return models

    def _check_changing_exog(self, changing_exog, n_samples, endog_names, discrete_endog_names):
        if changing_exog is None:
            return {}

        if not isinstance(changing_exog, dict):
            raise RuntimeError("changing_exog must be a dictionary.")

        changing_exog_ = {}
        for var_name, values in changing_exog.items():
            if var_name not in endog_names:
                raise RuntimeError(f"Unknown key in changing_exog. ({var_name})")

            if var_name in discrete_endog_names:
                raise RuntimeError(
                    f"Discrete variables shall not be specified. ({var_name})"
                )

            s = check_array(values, ensure_2d=False, dtype=None).ravel()
            if s.shape[0] != n_samples:
                raise RuntimeError(f"Wrong length. ({s.shape[0]} != {n_samples})")

            changing_exog_[var_name] = values

        return changing_exog_

    def _check_changing_models(self, changing_models, endog_names, discrete_endog_names):
        if changing_models is None:
            return {}

        if not isinstance(changing_models, dict):
            raise RuntimeError("changing_models shall be list.")

        changing_models_ = {}
        for target_name, model_info in changing_models.items():
            if not isinstance(model_info, dict):
                raise RuntimeError("changing_models shall be list of dictionaries.")

            # check target_name
            if not isinstance(target_name, str):
                raise TypeError("Key of changing_models must be str.")
            if target_name not in endog_names:
                raise RuntimeError(f"Unknown name. ({target_name})")

            # parent_names key
            if "parent_names" not in model_info.keys():
                raise KeyError("model_info must have parent_names key.")

            parent_names = model_info["parent_names"]
            if not isinstance(parent_names, list):
                raise KeyError("parent_names must be list.")

            for parent_name in model_info["parent_names"]:
                if parent_name not in endog_names:
                    raise RuntimeError(f"Unknown name. ({parent_name})")

            if len(parent_names) == 0:
                changing_models_[target_name] = {"parent_names": []}
                continue

            # coef and model key
            if "coef" not in model_info.keys() and "model" not in model_info.keys():
                raise KeyError("Elements of changing_models must have coef or model key when parent_names is set.")

            # coef key
            if "coef" in model_info.keys():
                coef = check_array(model_info["coef"], ensure_2d=False, ensure_min_samples=0)
                if len(coef) != len(parent_names):
                    raise ValueError("len(coef) != len(parent_names)")

                if target_name not in discrete_endog_names:
                    model = _LinearRegression(coef)
                else:
                    model = _LogisticRegression(coef)

                changing_models_[target_name] = {"parent_names": parent_names, "model": model}
                continue

            # model key
            if "model" not in model_info.keys() or model_info["model"] is None:
                raise KeyError("model must be set when coef isn't set.")

            model = model_info["model"]
            self._check_model_instance(model, target_name, discrete_endog_names)

            changing_models_[target_name] = {"parent_names": parent_names, "model": model}

        return changing_models_

    def _dispatch_impl(self, cd_algo_name):
        if cd_algo_name is None:
            return CBSILiNGAM
        elif cd_algo_name == DirectLiNGAM.__name__:
            return CBSILiNGAM
        elif cd_algo_name == ICALiNGAM.__name__:
            return CBSILiNGAM
        elif cd_algo_name == BottomUpParceLiNGAM.__name__:
            return CBSIUnobsCommonCauseLiNGAM
        elif cd_algo_name == VARLiNGAM.__name__:
            return CBSITimeSeriesLiNGAM
        else:
            raise ValueError("Unknown cd_algo_name")

    def _train(self, models, causal_graph, impl):
        train_result = {}

        for target_name in impl.endog_names_:
            y = impl.get_data(target_name)
            y = y.ravel()

            parent_names = impl.get_parent_names(target_name)
            if len(parent_names) == 0:
                train_result[target_name] = {
                    "model": None,
                    "parent_names": [],
                    "predicted": None,
                    "residual": y,
                }
                continue

            X = impl.get_data(parent_names)

            is_classifier = target_name in impl.discrete_endog_names_

            # select a model to train
            if target_name in models.keys():
                model = clone(models[target_name])
            else:
                if is_classifier:
                    model = LogisticRegression()
                else:
                    model = LinearRegression()

            model.fit(X, y)
            predicted = model.predict(X)
            predicted = predicted.ravel()

            # compute residuals
            if not is_classifier:
                residual = (y - predicted).ravel()
            else:
                residual = None

            train_result[target_name] = {
                "model": model,
                "parent_names": parent_names,
                "predicted": predicted,
                "residual": residual,
            }

        return train_result

    def _simulate(
        self,
        changing_exog,
        changing_models,
        impl,
        train_result,
        shuffle_residual,
        random_state,
    ):
        simulated = pd.DataFrame(index=np.arange(impl.exog_length_), columns=impl.endog_names_)

        if shuffle_residual:
            shuffle_index = random_state.choice(
                np.arange(impl.exog_length_),
                size=impl.exog_length_,
                replace=False
            )

        # modify causal order
        changing_edges = {}
        for target_name, info in changing_models.items():
            changing_edges[target_name] = info["parent_names"]
        causal_order = self._impl.get_causal_order(changing_edges)

        # predict from upstream to downstream
        for target_name in causal_order:
            is_classifier = target_name in impl.discrete_endog_names_

            # data
            parent_names = impl.get_parent_names(target_name)
            if target_name in changing_models.keys():
                parent_names_ = changing_models[target_name]["parent_names"]
                if parent_names_ is not None:
                    parent_names = parent_names_

            # error
            if target_name not in changing_exog.keys():
                error = train_result[target_name]["residual"]
            else:
                error = changing_exog[target_name].ravel()

            if shuffle_residual and error is not None:
                error = error[shuffle_index]

            # exogenous variable
            if len(parent_names) == 0:
                if not is_classifier:
                    simulated[target_name] = error
                else:
                    simulated[target_name] = impl.get_data(target_name)
                continue

            # model
            if target_name not in changing_models.keys():
                model = train_result[target_name]["model"]
            else:
                model = changing_models[target_name]["model"]

            # predict
            predicted = model.predict(simulated[parent_names].values)

            predicted = predicted.ravel()
            if not is_classifier:
                predicted = predicted + error

            simulated[target_name] = predicted

        return simulated


class CBSImpl(metaclass=ABCMeta):
    """ Class for handling a given causal_graph and data
    """

    def __init__(self, X, causal_graph, is_discrete=None):
        """ Constructor

        Constructor must expose properties below.
        """
        raise NotImplementedError

    @property
    def endog_names_(self):
        """ List of endogenous variable names
        Names of variables must be unique.
        """
        raise NotImplementedError

    @property
    def discrete_endog_names_(self):
        """ List of discrete endogenous variable names
        """
        raise NotImplementedError

    @property
    def causal_order_(self):
        """ List showing the causal order in a given causal graph.
        """
        raise NotImplementedError

    @property
    def exog_length_(self):
        """ Length of a estiamted residual data.
        """
        raise NotImplementedError

    @abstractmethod
    def get_parent_names(self, var_name):
        """ Return parent names of var_name.

        Parameters
        ----------
        var_name : str
            Target variable name.

        Returns
        -------
        parent_names : list
            List contains parent names of var_name.
        """
        raise NotImplementedError

    @abstractmethod
    def get_data(self, var_names):
        """ Return data of var_names

        Parameters
        ----------
        var_names : list
            List of target variable names.

        Returns
        -------
        data : numpy.ndarray
            Data of var_names.
        """
        raise NotImplementedError

    @abstractmethod
    def get_causal_order(self, changing_edges=None):
        """ Return a causal order in a given causal_graph and changing_edges

        Parameters
        ----------
        changing_edges : dict, default=None
            Dictaionry about edge modifications. Key is a target endogenous variable name and
            value is a list of parent names of the target endogenous variable.

        Returns
        -------
        causal_order : list
            List of endogenous variable names.
        """
        raise NotImplementedError


class CBSILiNGAM(CBSImpl):
    """ Class for data handling for ICALiNGAM and DirectLiNGAM. """

    def __init__(self, X, causal_graph, is_discrete=None):
        X_ = check_array(X)

        n_samples, n_features = X_.shape

        causal_graph = self._check_causal_graph(causal_graph, X_)

        endog_names, discrete_endog_names = self._make_var_names(causal_graph, X, is_discrete)

        causal_order = self._calc_causal_order(causal_graph)
        if causal_order is None:
            raise ValueError("causal_graph must be acyclic.")
        causal_order = [endog_names[n] for n in causal_order]

        self._X = X_
        self._exog_length = n_samples
        self._is_discrete = is_discrete
        self._endog_names = endog_names
        self._discrete_endog_names = discrete_endog_names
        self._causal_graph = causal_graph
        self._causal_order = causal_order

    @property
    def causal_order_(self):
        return self._causal_order

    @property
    def endog_names_(self):
        return self._endog_names

    @property
    def discrete_endog_names_(self):
        return self._discrete_endog_names

    @property
    def exog_length_(self):
        return self._exog_length

    def get_parent_names(self, var_name):
        if var_name not in self._endog_names:
            raise ValueError(f"Unknown variable. {var_name}")

        causal_graph = ~np.isclose(self._causal_graph, 0)

        index = self._endog_names.index(var_name)
        parent_indices = np.argwhere(causal_graph[index, :]).ravel()
        parent_names = np.array(self._endog_names)[parent_indices].tolist()

        return parent_names

    def get_data(self, var_names):
        if isinstance(var_names, str):
            var_names = [var_names]

        var_indices = []
        for var_name in var_names:
            index = self._endog_names.index(var_name)
            var_indices.append(index)
        data = self._X[:, var_indices]

        return data

    def get_causal_order(self, changing_edges=None):
        if changing_edges is None:
            changing_edges = {}

        causal_graph = self._causal_graph.copy()

        for target_name, parent_names in changing_edges.items():
            row = self._endog_names.index(target_name)
            cols = [self._endog_names.index(parent_name) for parent_name in parent_names]

            causal_graph[row, :] = 0
            causal_graph[row, cols] = 1

        causal_order = self._calc_causal_order(causal_graph)
        if causal_order is None:
            mes = "causal_graph updated by changing_models is cyclic."
            mes += " changing_models must be set so that causal graph does not cycle."
            raise ValueError(mes)
        causal_order = [self._endog_names[n] for n in causal_order]

        return causal_order

    def _check_causal_graph(self, causal_graph, X):
        try:
            causal_graph = check_array(causal_graph, copy=True)

            n_features = X.shape[1]
            if causal_graph.shape != (n_features, n_features):
                raise ValueError("The shape of causal_graph must be (n_features, n_features)")
        except Exception as e:
            raise ValueError("causal_graph has an error: " + str(e))

        return causal_graph

    def _make_var_names(self, causal_graph, X, is_discrete):
        n_features = len(causal_graph)

        if isinstance(X, pd.DataFrame):
            endog_names = check_array(X.columns, ensure_2d=False, dtype=None, copy=True)
            endog_names = [str(endog_name) for endog_name in endog_names]
        else:
            endog_names = [f"{i:d}" for i in range(n_features)]

        if is_discrete is None:
            is_discrete = [False for _ in range(n_features)]

        discrete_endog_names = np.array(endog_names)[is_discrete].tolist()

        return endog_names, discrete_endog_names

    def _calc_causal_order(self, causal_graph):
        """Obtain a causal order from the given causal_graph strictly.

        Parameters
        ----------
        causal_graph : array-like, shape (n_features, n_samples)
            Target causal_graph.

        Return
        ------
        causal_order : array, shape [n_features, ]
            A causal order of the given causal_graph on success, None otherwise.
        """
        causal_order = []

        row_num = causal_graph.shape[0]
        original_index = np.arange(row_num)

        while 0 < len(causal_graph):
            # find a row all of which elements are zero
            row_index_list = np.where(np.sum(np.abs(causal_graph), axis=1) == 0)[0]
            if len(row_index_list) == 0:
                break

            target_index = row_index_list[0]

            # append i to the end of the list
            causal_order.append(original_index[target_index])
            original_index = np.delete(original_index, target_index, axis=0)

            # remove the i-th row and the i-th column from causal_graph
            mask = np.delete(np.arange(len(causal_graph)), target_index, axis=0)
            causal_graph = causal_graph[mask][:, mask]

        if len(causal_order) != row_num:
            causal_order = None

        return causal_order


class CBSIUnobsCommonCauseLiNGAM(CBSILiNGAM):
    """ Class for data handling for BottomUpParceLiNGAM. """

    def _check_causal_graph(self, causal_graph, X):
        try:
            # fill nan with zeros
            causal_graph = check_array(causal_graph, force_all_finite="allow-nan", copy=True)

            n_features = X.shape[1]
            if causal_graph.shape != (n_features, n_features):
                raise RuntimeError("The shape of causal_graph must be (n_features, n_features)")

            causal_graph[np.isnan(causal_graph)] = 0
        except Exception as e:
            raise ValueError("causal_graph has an error: " + str(e))

        return causal_graph


class CBSITimeSeriesLiNGAM(CBSILiNGAM):
    """ Class for data handling for VARLiNGAM. """

    def __init__(self, X, causal_graph, is_discrete=None):
        super().__init__(X, causal_graph, is_discrete=is_discrete)

        # data is shortened to make lags
        self._exog_length = self._exog_length - self._n_lags

    def get_data(self, var_names):
        n_features = self._X.shape[1]

        if isinstance(var_names, str):
            var_names = [var_names]

        X_ = []
        for var_name in var_names:
            index = self._endog_names.index(var_name)
            X_index = index % n_features
            lag = index // n_features

            data = self._X[self._n_lags - lag:-lag if lag != 0 else None, X_index]

            X_.append(data)
        X_ = np.vstack(X_).T

        return X_

    def _check_causal_graph(self, causal_graph, X):
        try:
            causal_graph = check_array(causal_graph, allow_nd=True, copy=True)
            if len(causal_graph.shape) != 3:
                raise ValueError("causal_graph must be 3 dimentional array.")

            n_features = X.shape[1]
            if causal_graph.shape[1:] != (n_features, n_features):
                raise ValueError("The shape of causal_graph must be (n_lags+1, n_features, n_features)")
        except Exception as e:
            raise ValueError("causal_graph has an error: " + str(e))

        coef = np.concatenate(causal_graph, axis=1)

        causal_graph_ = np.zeros((coef.shape[1], coef.shape[1]))
        causal_graph_[:len(coef)] = coef

        self._n_lags = len(causal_graph) - 1

        return causal_graph_

    def _make_var_names(self, causal_graph, X, is_discrete):
        n_features = X.shape[1]

        if isinstance(X, pd.DataFrame):
            endog_names_ = X.columns.tolist()
        else:
            endog_names_ = [f"{i:d}" for i in range(n_features)]

        endog_names = []
        for i in range(self._n_lags + 1):
            if i == 0:
                index_format = "[t]"
            else:
                index_format = f"[t-{i}]"
            endog_names += [name + index_format for name in endog_names_]

        if is_discrete is None:
            is_discrete = [False for _ in range(n_features)]
            is_discrete *= self._n_lags + 1
        else:
            is_discrete = is_discrete * (self._n_lags + 1)

        discrete_endog_names = np.array(endog_names)[is_discrete].tolist()

        return endog_names, discrete_endog_names


class _LinearRegression():
    """ Linear regression model with configurable coefficients """

    def __init__(self, coef):
        self._coef = np.array(coef)

    def predict(self, X):
        if len(self._coef) == 0:
            return np.zeros((len(X), 1))
        return (self._coef @ X.T).T

    @property
    def coef_(self):
        return self._coef


class _LogisticRegression():
    """ Logistic regression model with configurable coefficients """

    def __init__(self, coef):
        self._coef = np.array(coef)

    def predict(self, X):
        if len(self._coef) == 0:
            return np.zeros((len(X), 1))
        return expit((self._coef @ X.T).T) >= 0.5

    @property
    def coef_(self):
        return self._coef



import os
import random

import numpy as np
import pandas as pd

from lingam import MultiGroupRCD


def test_fit_success():
    def get_coef():
        coef = random.random()
        return coef if coef >= 0.5 else coef - 1.0

    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3

    B1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 500
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B1[0, 5] + get_external_effect(samples)
    x1 = x5 * B1[1, 5] + get_external_effect(samples)
    x2 = x0 * B1[2, 0] + x1 * B1[2, 1] + get_external_effect(samples)
    x3 = x2 * B1[3, 2] + x6 * B1[3, 6] + get_external_effect(samples)
    x4 = x2 * B1[4, 2] + x6 * B1[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X1 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    B2 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 1000
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B2[0, 5] + get_external_effect(samples)
    x1 = x5 * B2[1, 5] + get_external_effect(samples)
    x2 = x0 * B2[2, 0] + x1 * B2[2, 1] + get_external_effect(samples)
    x3 = x2 * B2[3, 2] + x6 * B2[3, 6] + get_external_effect(samples)
    x4 = x2 * B2[4, 2] + x6 * B2[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X2 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    X_list = [X1, X2]

    model = MultiGroupRCD(
        max_explanatory_num=2,
        cor_alpha=0.01,
        ind_alpha=0.01,
        shapiro_alpha=0.01,
        MLHSICR=True,
        bw_method="mdbs",
    )

    model.fit(X_list)

    effects = model.estimate_total_effect(X_list, from_index=2, to_index=4)
    p_values = model.get_error_independence_p_values(X_list)
    model.adjacency_matrices_
    model.ancestors_list_

    # f-correlation
    model = MultiGroupRCD(
        independence="fcorr",
    )
    model.fit(X_list)


def test_fit_invalid_data():
    def get_coef():
        coef = random.random()
        return coef if coef >= 0.5 else coef - 1.0

    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3

    B1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 500
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B1[0, 5] + get_external_effect(samples)
    x1 = x5 * B1[1, 5] + get_external_effect(samples)
    x2 = x0 * B1[2, 0] + x1 * B1[2, 1] + get_external_effect(samples)
    x3 = x2 * B1[3, 2] + x6 * B1[3, 6] + get_external_effect(samples)
    x4 = x2 * B1[4, 2] + x6 * B1[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X1 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    B2 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 500
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B2[0, 5] + get_external_effect(samples)
    x1 = x5 * B2[1, 5] + get_external_effect(samples)
    x2 = x0 * B2[2, 0] + x1 * B2[2, 1] + get_external_effect(samples)
    x3 = x2 * B2[3, 2] + x6 * B2[3, 6] + get_external_effect(samples)
    x4 = x2 * B2[4, 2] + x6 * B2[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X2 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    X_list = [X1, X2]

    # Not array data
    X = 1
    try:
        model = MultiGroupRCD()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: max_explanatory_num
    try:
        model = MultiGroupRCD(max_explanatory_num=-1)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: cor_alpha
    try:
        model = MultiGroupRCD(cor_alpha=-1)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: ind_alpha
    try:
        model = MultiGroupRCD(ind_alpha=-1)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: shapiro_alpha
    try:
        model = MultiGroupRCD(shapiro_alpha=-1)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: bw_method
    try:
        model = MultiGroupRCD(bw_method="X")
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: independence
    try:
        model = MultiGroupRCD(independence="lingam")
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid value: ind_corr
    try:
        model = MultiGroupRCD(ind_corr=-0.5)
        model.fit(X_list)
    except ValueError:
        pass
    else:
        raise AssertionError


def test_bootstrap_success():
    def get_coef():
        coef = random.random()
        return coef if coef >= 0.5 else coef - 1.0

    get_external_effect = lambda n: np.random.normal(0.0, 0.5, n) ** 3

    B1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 500
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B1[0, 5] + get_external_effect(samples)
    x1 = x5 * B1[1, 5] + get_external_effect(samples)
    x2 = x0 * B1[2, 0] + x1 * B1[2, 1] + get_external_effect(samples)
    x3 = x2 * B1[3, 2] + x6 * B1[3, 6] + get_external_effect(samples)
    x4 = x2 * B1[4, 2] + x6 * B1[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X1 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    B2 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, get_coef(), 0.0],
            [get_coef(), get_coef(), 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, get_coef(), 0.0, 0.0, 0.0, get_coef()],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ]
    )

    samples = 500
    x5 = get_external_effect(samples)
    x6 = get_external_effect(samples)
    x0 = x5 * B2[0, 5] + get_external_effect(samples)
    x1 = x5 * B2[1, 5] + get_external_effect(samples)
    x2 = x0 * B2[2, 0] + x1 * B2[2, 1] + get_external_effect(samples)
    x3 = x2 * B2[3, 2] + x6 * B2[3, 6] + get_external_effect(samples)
    x4 = x2 * B2[4, 2] + x6 * B2[4, 6] + get_external_effect(samples)

    # x5, x6 is a latent variable.
    X2 = pd.DataFrame(
        np.array([x0, x1, x2, x3, x4, x5]).T,
        columns=["x0", "x1", "x2", "x3", "x4", "x5"],
    )

    X_list = [X1, X2]

    model = MultiGroupRCD(
        max_explanatory_num=2,
        cor_alpha=0.01,
        ind_alpha=0.01,
        shapiro_alpha=0.01,
        MLHSICR=True,
        bw_method="mdbs",
    )

    results = model.bootstrap(X_list, n_sampling=2)
    cdc = results[0].get_causal_direction_counts(n_directions=8, min_causal_effect=0.01)
    dagc = results[0].get_directed_acyclic_graph_counts(
        n_dags=3, min_causal_effect=0.01
    )
    prob = results[0].get_probabilities(min_causal_effect=0.01)
    effects = results[0].get_total_causal_effects(min_causal_effect=0.01)
    paths = results[0].get_paths(from_index=3, to_index=1)



import os

import numpy as np
import pandas as pd
from lingam.group_lingam import GroupLiNGAM

def get_external_effect(n_samples):
    zi = np.random.normal(0, 1, n_samples)
    qi = np.random.choice(np.concatenate((np.random.uniform(0.5, 0.8, n_samples//2),
                                          np.random.uniform(1.2, 2.0, n_samples//2))))
    ei = np.sign(zi) * np.abs(zi) ** qi
    ei = (ei - np.mean(ei)) / np.std(ei)
    return ei

def test_fit_success():
    # causal direction: x0, x1 --> x2, x3 --> x4
    x0 = get_external_effect(500)
    x1 = get_external_effect(500)
    x2 = -1.21 * x1 + get_external_effect(500)
    x3 = -1.98 * x0 + get_external_effect(500)
    x4 = -0.55 * x1 - 0.50 * x2 - 1.16 * x3 + get_external_effect(500)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = GroupLiNGAM()
    model.fit(X)

    co = model.causal_order_
    am = model.adjacency_matrix_

def test_fit_invalid():
    # Not array data
    X = 1
    try:
        model = GroupLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = GroupLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = GroupLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = GroupLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

def test_bootstrap_success():
    # causal direction: x0, x1 --> x2, x3 --> x4
    x0 = get_external_effect(500)
    x1 = get_external_effect(500)
    x2 = -1.21 * x1 + get_external_effect(500)
    x3 = -1.98 * x0 + get_external_effect(500)
    x4 = -0.55 * x1 - 0.50 * x2 - 1.16 * x3 + get_external_effect(500)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    model = GroupLiNGAM()
    result = model.bootstrap(X, n_sampling=5)
    am = result.adjacency_matrices_
    assert len(am) == 5
    te = result.total_effects_
    assert len(te) == 5

    # No argument
    cdc = result.get_causal_direction_counts()

    # n_directions=2
    cdc = result.get_causal_direction_counts(n_directions=2)

    # min_causal_effect=0.2
    cdc = result.get_causal_direction_counts(min_causal_effect=0.2)

    # split_by_causal_effect_sign=True
    cdc = result.get_causal_direction_counts(split_by_causal_effect_sign=True)

    # No argument
    dagc = result.get_directed_acyclic_graph_counts()

    # n_dags=2
    dagc = result.get_directed_acyclic_graph_counts(n_dags=2)

    # min_causal_effect=0.6
    dagc = result.get_directed_acyclic_graph_counts(min_causal_effect=0.6)

    # split_by_causal_effect_sign=True
    dagc = result.get_directed_acyclic_graph_counts(split_by_causal_effect_sign=True)

    # get_probabilities
    probs = result.get_probabilities()

    # get_probabilities
    probs = result.get_probabilities(min_causal_effect=0.6)

    # get_total_causal_effects
    ce = result.get_total_causal_effects()

    # get_total_causal_effects
    ce = result.get_total_causal_effects(min_causal_effect=0.6)

    # get_paths
    paths = result.get_paths(1, 2)

def test_bootstrap_invalid_data():
    # causal direction: x0, x1 --> x2, x3 --> x4
    x0 = get_external_effect(500)
    x1 = get_external_effect(500)
    x2 = -1.21 * x1 + get_external_effect(500)
    x3 = -1.98 * x0 + get_external_effect(500)
    x4 = -0.55 * x1 - 0.50 * x2 - 1.16 * x3 + get_external_effect(500)
    X = pd.DataFrame(np.array([x0, x1, x2, x3, x4]).T, columns=['x0', 'x1', 'x2', 'x3', 'x4'])

    # Invalid argument: bootstrap(n_sampling=-1)
    model = GroupLiNGAM()
    try:
        result = model.bootstrap(X, n_sampling=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: bootstrap(n_sampling='3')
    model = GroupLiNGAM()
    try:
        result = model.bootstrap(X, n_sampling='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions=-1)
    model = GroupLiNGAM()
    result = model.bootstrap(X, n_sampling=5)
    try:
        result.get_causal_direction_counts(n_directions=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(n_directions='3')
    try:
        result.get_causal_direction_counts(n_directions='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_causal_direction_counts(min_causal_effect=-1.0)
    try:
        result.get_causal_direction_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags=-1)
    try:
        result.get_directed_acyclic_graph_counts(n_dags=-1)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(n_dags='3')
    try:
        result.get_directed_acyclic_graph_counts(n_dags='3')
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    try:
        result.get_directed_acyclic_graph_counts(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_probabilities(min_causal_effect=-1.0)
    try:
        result.get_probabilities(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_total_causal_effects(min_causal_effect=-1.0)
    try:
        result.get_total_causal_effects(min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Invalid argument: get_paths(min_causal_effect=-1.0)
    try:
        result.get_paths(0, 1, min_causal_effect=-1.0)
    except ValueError:
        pass
    else:
        raise AssertionError



"""
The lingam module includes implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

from .bootstrap import BootstrapResult
from .bottom_up_parce_lingam import BottomUpParceLiNGAM
from .camuv import CAMUV
from .causal_based_simulator import CausalBasedSimulator
from .causal_effect import CausalEffect
from .direct_lingam import DirectLiNGAM
from .ica_lingam import ICALiNGAM
from .lim import LiM
from .longitudinal_lingam import LongitudinalBootstrapResult, LongitudinalLiNGAM
from .multi_group_direct_lingam import MultiGroupDirectLiNGAM
from .multi_group_rcd import MultiGroupRCD
from .rcd import RCD
from .resit import RESIT
from .var_lingam import VARBootstrapResult, VARLiNGAM
from .varma_lingam import VARMABootstrapResult, VARMALiNGAM
from .lina import LiNA
from .lina import MDLiNA
from .high_dim_direct_lingam import HighDimDirectLiNGAM
from .multi_group_camuv import MultiGroupCAMUV
from .group_direct_lingam import GroupDirectLiNGAM
from .group_lingam import GroupLiNGAM

__all__ = [
    "ICALiNGAM",
    "DirectLiNGAM",
    "BootstrapResult",
    "MultiGroupDirectLiNGAM",
    "CausalEffect",
    "VARLiNGAM",
    "VARMALiNGAM",
    "LongitudinalLiNGAM",
    "VARBootstrapResult",
    "VARMABootstrapResult",
    "LongitudinalBootstrapResult",
    "BottomUpParceLiNGAM",
    "RCD",
    "CAMUV",
    "RESIT",
    "LiM",
    "CausalBasedSimulator",
    "MultiGroupRCD",
    "LiNA",
    "MDLiNA",
    "HighDimDirectLiNGAM",
    "MultiGroupCAMUV",
    "GroupDirectLiNGAM",
    "GroupLiNGAM",
]

__version__ = "1.11.0"
